          ///////////////////////////////////////////////////////////////////////
          //                                                                   //
          //       SoftSpot Software Component Library                         //
          //       Key Maker software component                                //
          //                                                                   //
          //       Copyright (c) 1996 - 2002 SoftSpot Software Ltd.            //
          //       ALL RIGHTS RESERVED                                         //
          //                                                                   //
          //   The entire contents of this file is protected by U.S. and       //
          //   International Copyright Laws. Unauthorized reproduction,        //
          //   reverse-engineering, and distribution of all or any portion of  //
          //   the code contained in this file is strictly prohibited and may  //
          //   result in severe civil and criminal penalties and will be       //
          //   prosecuted to the maximum extent possible under the law.        //
          //                                                                   //
          //   RESTRICTIONS                                                    //
          //                                                                   //
          //   THE SOURCE CODE CONTAINED WITHIN THIS FILE AND ALL RELATED      //
          //   FILES OR ANY PORTION OF ITS CONTENTS SHALL AT NO TIME BE        //
          //   COPIED, TRANSFERRED, SOLD, DISTRIBUTED, OR OTHERWISE MADE       //
          //   AVAILABLE TO OTHER INDIVIDUALS WITHOUT EXPRESS WRITTEN CONSENT  //
          //   AND PERMISSION FROM SOFTSPOT SOFTWARE LTD.                      //
          //                                                                   //
          //   CONSULT THE END USER LICENSE AGREEMENT FOR INFORMATION ON       //
          //   ADDITIONAL RESTRICTIONS.                                        //
          //                                                                   //
          ///////////////////////////////////////////////////////////////////////

unit ajFileBrowser;

{                               -=< Key Maker File Browser >=-
{
{ Copyright SoftSpot Software 2002 - All Rights Reserved
{
{ Author        : Andrew J Jameson
{
{ Web site      : www.softspotsoftware.com
{ e-mail        : contact@softspotsoftware.com
{
{ Creation Date : 01 March 2002
{
{ Version       : 1.00
{
{ Description   : File property editor for Key Maker.                                              }

interface


uses
  Forms, Windows, ImgList, Controls, Menus, ComCtrls, StdCtrls, ExtCtrls,
  Graphics, Buttons, Classes, Messages, ajKeyMakerAbout, ajAlterMenu, ajRecentKeys;

type
  TajColumnSettings = record
    DefWidth  : integer;
    Visible   : boolean;
  end;

type
  TfrmFileBrowser = class(TForm)
    Bevel1            : TBevel;
    Bevel2            : TBevel;
    btnBookmark       : TSpeedButton;
    btnCancel         : TSpeedButton;
    btnDetails        : TSpeedButton;
    btnLargeIcons     : TSpeedButton;
    btnList           : TSpeedButton;
    btnRefresh        : TSpeedButton;
    btnSmallIcons     : TSpeedButton;
    btnUpdate         : TSpeedButton;
    edFilename        : TEdit;
    imButtons         : TImageList;
    imLarge           : TImageList;
    imMenu            : TImageList;
    imSmall           : TImageList;
    lvDrives          : TListView;
    lvFiles           : TListView;
    lvFilesPopupMenu  : TPopupMenu;
    MainMenu          : TMainMenu;
    mnuAbout          : TMenuItem;
    mnuAccessedCol    : TMenuItem;
    mnuArchive        : TMenuItem;
    mnuAttributes     : TMenuItem;
    mnuAttributesCol  : TMenuItem;
    mnuBookmarks      : TMenuItem;
    mnuColumns        : TMenuItem;
    mnuCopyPath       : TMenuItem;
    mnuCreatedCol     : TMenuItem;
    mnuDetails        : TMenuItem;
    mnuEditFileParams : TMenuItem;
    mnuExit           : TMenuItem;
    mnuExplorer       : TMenuItem;
    mnuHidden         : TMenuItem;
    mnuLargeIcons     : TMenuItem;
    mnuList           : TMenuItem;
    mnuModifiedCol    : TMenuItem;
    mnuNormal         : TMenuItem;
    mnuOptions        : TMenuItem;
    mnuPopBookmark    : TMenuItem;
    mnuReadOnly       : TMenuItem;
    mnuRefresh        : TMenuItem;
    mnuShowFolders    : TMenuItem;
    mnuSizeCol        : TMenuItem;
    mnuSmallIcons     : TMenuItem;
    mnuSystem         : TMenuItem;
    mnuTypeCol        : TMenuItem;
    mnuUpdate         : TMenuItem;
    mnuView           : TMenuItem;
    mnuViewType       : TMenuItem;
    mnuXPMenu         : TMenuItem;
    N1                : TMenuItem;
    N2                : TMenuItem;
    N5                : TMenuItem;
    N6                : TMenuItem;
    pnBack            : TPanel;
    pnFilename        : TPanel;
    pnToolbar         : TPanel;
    StatusBar         : TStatusBar;
    procedure FormCreate              (Sender : TObject);
    procedure FormShow                (Sender : TObject);
    procedure FormResize              (Sender : TObject);    
    procedure FormClose               (Sender : TObject; var Action : TCloseAction);
    procedure FormDestroy             (Sender : TObject);
    procedure mnuAboutClick           (Sender : TObject);
    procedure mnuAttributesClick      (Sender : TObject);
    procedure mnuBookmarkClick        (Sender : TObject);
    procedure mnuColumnsClick         (Sender : TObject);
    procedure mnuCopyPathClick        (Sender : TObject);
    procedure mnuEditFileParamsClick  (Sender : TObject);
    procedure mnuExitClick            (Sender : TObject);
    procedure mnuShowFoldersClick     (Sender : TObject);
    procedure mnuRefreshClick         (Sender : TObject);
    procedure mnuUpdateClick          (Sender : TObject);
    procedure mnuViewTypeClick        (Sender : TObject);
    procedure mnuXPMenuClick          (Sender : TObject);
    procedure lvFilesColumnClick      (Sender : TObject; Column : TListColumn);
    procedure lvFilesCompare          (Sender : TObject; Item1, Item2 : TListItem; Data : integer; var Compare : integer);
    procedure lvFilesDblClick         (Sender : TObject);
    procedure lvFilesKeyDown          (Sender : TObject; var Key : Word; Shift : TShiftState);
    procedure lvFilesSelectItem       (Sender : TObject; Item : TListItem; Selected : boolean);
    procedure lvDrivesCompare         (Sender : TObject; Item1, Item2 : TListItem; Data : integer; var Compare : integer);
    procedure lvDrivesDblClick        (Sender : TObject);
    procedure lvDrivesKeyDown         (Sender : TObject; var Key : Word; Shift : TShiftState);
  private
    { Private declarations }
    fSortArrows       : array[0..1] of TBitmap;
    fajAlterMenu      : TajAlterMenu;
    fajRecentKeys     : TajRecentKeys;
    fColumnSettings   : array[0..6] of TajColumnSettings;
    fFoldersHidden    : boolean;
    fIsPropertyEditor : boolean;
    fFileMask         : string;
    fFileAttributes   : DWORD;
    fSortAscending    : boolean;
    fColumnToSort     : integer;
    flvOldWindowProc  : TWndMethod;
    fRootIconIndex    : integer;
    fFilename         : string;
    procedure lvNewWindowProc         (var Message : TMessage);
  protected
    procedure SetfIsPropertyEditor    (Value : boolean);
    procedure SetfFilename            (Value : string);
    procedure GetSystemIcons;
    procedure SetFileAttributesFilter;
    procedure AddDrives;
    procedure RefreshlvFiles;    
    procedure RefreshlvDrives;
    procedure UpdateFileListView      (FileMask : string; Attribute : DWORD);
    procedure FileSelected            (Item : TListItem);
    procedure SetColumnBitmap         (Column : TListColumn);
    procedure RecentKeyClick          (KeyName : string);
    procedure SetBookmarkControls     (Tag : integer);
    function  FileFindAndSelect       (ListView : TListView; FileCaption : string) : boolean;
  public
    { Public declarations }
    property  IsPropertyEditor  : boolean   read fIsPropertyEditor  write SetfIsPropertyEditor;
    property  Filename          : string    read fFilename          write SetfFilename;
  end;

var
  frmFileBrowser : TfrmFileBrowser;

implementation

{$R *.DFM}
{$R FileBrowser.res}

{$IFDEF VER140}   // Delphi6
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}

uses
  SysUtils, CommCtrl, ShellAPI, Registry, Clipbrd, ajFileEditParams;

const
  cColName        = 0;                                                    // Column references.
  cColSize        = 1;
  cColType        = 2;
  cColCreated     = 3;
  cColModified    = 4;
  cColAccessed    = 5;
  cColAttributes  = 6;
  cColFullName    = 7;
  cColDirFile     = 8;
  cColDriveName   = 5;

  cAddBookmark    = 3;
  cRemoveBookmark = 4;

type
  TajRegReadWrite = (tRegRead, tRegWrite);

{--------------------------------------------------------------------------------------------------}
{                                 Registry - Read / Write Menu                                     }
{--------------------------------------------------------------------------------------------------}

procedure RegSubMenuSettings(Menu : TMenuItem; RegPath : string; ReadWrite : TajRegReadWrite);
var
  lp1   : integer;
  Name  : string;
begin
  with TRegistry.Create do begin
    try
      RootKey := cRegHKEY;
      if OpenKey(RegPath, ReadWrite = tRegWrite) then begin
        for lp1 := 0 to pred(Menu.Count) do begin
          if (Menu.Items[lp1].Count > 0) then
            RegSubMenuSettings(Menu.Items[lp1], RegPath, ReadWrite)
          else begin
            Name  := StripHotKey(Menu.Items[lp1].Caption);
            if (Name <> '-') then begin
              if (ReadWrite = tRegRead) then begin
                if ValueExists(Name) then begin
                  if (Menu.Items[lp1].Checked <> ReadBool(Name)) then
                    Menu.Items[lp1].Click;
                end; {if}
              end else
                WriteBool(Name, Menu.Items[lp1].Checked);
            end; {if}
          end; {if}
        end; {for}
        CloseKey;
      end; {if}
    finally;
      Free;
    end; {try}
  end; {with}
end; {RegSubMenuSettings}

{--------------------------------------------------------------------------------------------------}
{                             Miscellaneous Procedure & Functions                                  }
{--------------------------------------------------------------------------------------------------}

type
  TDiskInfo = record
    MediaType     : string[10];
    Serial        : DWORD;
    Capacity      : Int64;
    FreeSpace     : Int64;
    VolumeLabel   : array[0..$FF] of char;
    FileSystem    : array[0..$FF] of char;
  end;

function GetDiskInfo(Value : string) : TDiskInfo;
var
  ErrorMode : Word;
  DW        : DWORD;
  SysFlags  : DWORD;
begin
  FillMemory(@Result, SizeOf(Result), 0);
  with Result do begin
    ErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
    try
      case GetDriveType(PChar(Value)) of
        DRIVE_UNKNOWN     : MediaType := 'Unknown';
        DRIVE_NO_ROOT_DIR : MediaType := 'NotExists';
        DRIVE_REMOVABLE   : MediaType := 'Removable';
        DRIVE_FIXED       : MediaType := 'Fixed';
        DRIVE_REMOTE      : MediaType := 'Remote';
        DRIVE_CDROM       : MediaType := 'CDROM';
        DRIVE_RAMDISK     : MediaType := 'RAMDisk';
      end; {case}
      if (Value <> 'A:\') then begin
        FreeSpace := DiskFree(ord(Value[1])-$40);
        Capacity  := DiskSize(ord(Value[1])-$40);
        GetVolumeInformation(PChar(Value), VolumeLabel, SizeOf(VolumeLabel), @Serial, DW, SysFlags, FileSystem, SizeOf(FileSystem));
      end; {if}
    finally
      SetErrorMode(ErrorMode);
    end; {try}
  end; {with}
end; {GetDiskInfo}

{--------------------------------------------------------------------------------------------------}

procedure HighlightFocusedItem(ListView : TListView);
// Highlights a focused item in a list view - if no item focused then selects the first item.
var
  Item  : TListItem;
begin
  if ListView.Visible then begin                                          // Can't do anything if it's not visible.
    with ListView do begin
      Item  := ItemFocused;                                               // Get the focused item ... and
      if (Item = nil) and (Items.Count > 0) then                          // if there isn't one and there are items
        Item  := Items[0];                                                // then pick the first one.
      if (Item <> nil) then begin                                         // If we've now got a valid item ...
        SetFocus;                                                         // Set focus on me.
        ItemFocused   := Item;                                            // Focus the item and
        Selected      := Item;                                            // select it too.
        Item.MakeVisible(true);                                           // Make sure that you can see me !
      end; {if}
    end; {with}
  end; {if}
end; {HighlightFocusedItem}

{--------------------------------------------------------------------------------------------------}
{                                       TfrmFileBrowser                                            }
{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.FormCreate(Sender : TObject);
var
  lp1 : integer;
begin
  for lp1 := 0 to 1 do begin
    fSortArrows[lp1]  := TBitmap.Create;
    case lp1 of
      0 : fSortArrows[lp1].LoadFromResourceName(hInstance, 'DOWNBITMAP');
      1 : fSortArrows[lp1].LoadFromResourceName(hInstance, 'UPBITMAP'  );
    end; {case}
  end; {for}

  pnBack.DoubleBuffered           := true;                                // Stop some of the flicker.
  pnFilename.DoubleBuffered       := true;
  edFilename.DoubleBuffered       := true;

  fajAlterMenu  := TajAlterMenu.Create(Self);                             // Create the XP menu extension object.
  fajRecentKeys                   := TajRecentKeys.Create(mnuBookmarks, fajAlterMenu, cRegFileBrowser);  // Add bookmark support.
  fajRecentKeys.OnRecentKeyClick  := RecentKeyClick;                      // Connect the event handler.
  Width         := 632;                                                   // The form kept getting resized in the
  Height        := 376;                                                   // IDE ... so set it what seems right !

  btnBookmark.Tag                 := cAddBookmark;                        // Initialize bookmark Tags.
  mnuPopBookmark.Tag              := cAddBookmark;

  for lp1 := cColName to cColAttributes do begin                          // Initialize the store for making
    fColumnSettings[lp1].DefWidth := lvFiles.Columns[lp1].Width;          // the columns visible.
    fColumnSettings[lp1].Visible  := true;
  end; {for}

  flvOldWindowProc      := lvFiles.WindowProc;                            // Exchange old window proc for new.
  lvFiles.WindowProc    := lvNewWindowProc;                               // All this to control column visibility !
  fFoldersHidden        := not mnuShowFolders.Checked;                    // Decide if folders are on view.
  fSortAscending        := true;                                          // Ascending sort.
  fColumnToSort         := cColName;                                      // Default - sort on name.
  IsPropertyEditor      := false;                                         // Not a property editor - might be later.
  lvDrives.LargeImages  := TImageList.Create(Self);                       // Create the image lists for the
  lvDrives.SmallImages  := TImageList.Create(Self);                       // system icons.
  lvFiles.LargeImages   := lvDrives.LargeImages;
  lvFiles.SmallImages   := lvDrives.SmallImages;
  GetSystemIcons;                                                         // Now go and get them !
  AddDrives;
  SetFileAttributesFilter;                                                // Set the file attributes filter.
  fFilename             := '';
  SetColumnBitmap(lvFiles.Column[0]);
end; {FormCreate}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.FormShow(Sender : TObject);
begin

  RegSubMenuSettings(mnuView, cRegPath + cRegFileBrowser, tRegRead);

  if lvDrives.Visible then
    HighlightFocusedItem(lvDrives);

  if lvFiles.Visible then
    FileFindAndSelect(lvFiles, ExtractFileName(fFilename));

  if not fIsPropertyEditor then
    SetBounds((Screen.Width - Width) div 2, (Screen.Height - Height) div 2, Width, Height);

  btnCancel.Left    := pnToolbar.Width - btnCancel.Width - 3;
  btnUpdate.Left    := btnCancel.Left  - btnUpdate.Width - 3;
  edFilename.Width  := pnFilename.Width - 8;

end; {FormShow}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.FormResize(Sender : TObject);
begin
  lvDrives.Arrange(arAlignTop);
  lvFiles.Arrange(arAlignTop);
end; {FormResize}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.FormClose(Sender : TObject; var Action : TCloseAction);
begin
  RegSubMenuSettings(mnuView, cRegPath + cRegFileBrowser, tRegWrite);
end; {FormClose}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.FormDestroy(Sender : TObject);
var
  lp1 : integer;
begin
  for lp1 := 0 to 1 do
    fSortArrows[lp1].Free;
  fajAlterMenu.Free;                                                      // Destroy the XP menu object.
  fajRecentKeys.Free;                                                     // Free bookmark support.
  lvFiles.WindowProc  := flvOldWindowProc;                                // Restore the old window proc ... blows up if we forget !
end; {FormDestroy}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvNewWindowProc(var Message : TMessage);        // All this to control the columns.
var
  hdn : ^THDNotify;                                                       // Notifcation parameter block.
begin
  if (Message.Msg = WM_NOTIFY) then begin                                 // Hey this could be for us !
    hdn := Pointer(Message.lParam);
    if (hdn.Item in [cColName..cColAttributes]) then begin                // One of our columns ?
      case  hdn.Hdr.code of                                               // If visible - then obey the dragging.
        HDN_BeginTrackA, HDN_BeginTrackW  : if fColumnSettings[hdn.Item].Visible then
                                              flvOldWindowProc(Message)
                                            else
                                              Message.Result  := 1;
        HDN_EndTrackA, HDN_EndTrackW      : begin                         // Update our width store.
                                              fColumnSettings[hdn.Item].DefWidth  := hdn.PItem.cxy;
                                              flvOldWindowProc(Message);
                                            end;
      else
        flvOldWindowProc(Message);                                        // Not for us after all - so pass it on.
      end; {case}
    end else
      flvOldWindowProc(Message);                                          // Neither was this one for us !
  end else
    flvOldWindowProc(Message);
end; {lvNewWindowProc}                                                    // Or this one either !


{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.SetfIsPropertyEditor(Value : boolean);
var
  lp1 : integer;
begin
  fIsPropertyEditor := Value;                                             // Hide or reveal stuff that
  mnuUpdate.Visible := fIsPropertyEditor;                                 // relates to the usage as a
  for lp1 := 0 to pred(pnToolbar.ControlCount) do                         // property editor.
    pnToolBar.Controls[lp1].Visible := (pnToolBar.Controls[lp1].Tag <> -1) or fIsPropertyEditor;
end; {SetfIsPropertyEditor}                                               // Uses tag of -1 for property editor.

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.SetfFilename(Value : string);
begin
  if (fFilename <> Value) then begin
    fFilename       := Value;
    edFilename.Text := fFilename;                                         // Display the full file name.
    UpdateFileListView(fFileMask, fFileAttributes);
  end; {if}
end; {SetfFilename}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.GetSystemIcons;
var
  SysImageList  : THandle;
  SHFileInfo    : TSHFileInfo;
begin
  with lvFiles do begin                                                   // What else - just get the icons.
    SysImageList  := SHGetFileInfo('', 0, SHFileInfo, SizeOf(TSHFileInfo), SHGFI_SYSICONINDEX or SHGFI_LARGEICON);
    if (SysImageList <> 0) then begin                                     // Large ones ...
      LargeImages.Handle      := SysImageList;
      LargeImages.AddImages(imLarge);                                     // Add the icon for the parent root directory.
      LargeImages.ShareImages := true;
    end; {if}
    SysImageList  := SHGetFileInfo('', 0, SHFileInfo, SizeOf(TSHFileInfo), SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
    if (SysImageList <> 0) then begin                                     // Small ones ...
      SmallImages.Handle      := SysImageList;
      SmallImages.AddImages(imSmall);                                     // Add the icon for the parent root directory.
      SmallImages.ShareImages := true;
    end; {if}
    fRootIconIndex  := pred(SmallImages.Count);
  end; {with}
end; {GetSystemIcons}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.SetFileAttributesFilter;
var
  lp1 : integer;
begin
  fFileAttributes := 0;                                                   // Build up the filter mask ...
  for lp1 := 0 to pred(mnuAttributes.Count) do                            // Note that the menu tags hold
    if mnuAttributes.Items[lp1].Checked then                              // the actual masks.
      fFileAttributes := fFileAttributes or DWORD(mnuAttributes.Items[lp1].Tag);
end; {SetFileAttributesFilter}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.AddDrives;
var
  Drives  : set of 0..25;
  ShInfo  : TSHFileInfo;
  lp1     : integer;
  Drv     : string;
  DI      : TDiskInfo;
begin
  lvDrives.Items.BeginUpdate;
  lvDrives.Items.Clear;
  Integer(Drives) := GetLogicalDrives;
  for lp1 := 0 to 25 do begin
    if (lp1 in Drives) then begin
      with lvDrives.Items.Add do begin
        Drv := chr(ord('A') + lp1) + ':\';
        try
          SHGetFileInfo(PChar(Drv), 0, ShInfo, SizeOf(ShInfo), SHGFI_SYSICONINDEX or SHGFI_DISPLAYNAME or SHGFI_TYPENAME);
    	    if (lvDrives.SmallImages <> nil) then
     	      ImageIndex  := ShInfo.iIcon;
          Caption := StrPas(ShInfo.szDisplayName);
          DI      := GetDiskInfo(Drv);
          SubItems.Add(DI.MediaType);
          if (DI.Serial = 0) then begin
            SubItems.Add('');
            SubItems.Add('');
            SubItems.Add('');
            SubItems.Add('');
          end else begin
            SubItems.Add(DI.FileSystem);
            SubItems.Add(FormatFloat('###,###,##0 KB', DI.Capacity  div 1024));
            SubItems.Add(FormatFloat('###,###,##0 KB', DI.FreeSpace div 1024));
            SubItems.Add(IntToHex(HiWord(DI.Serial), 4) + '-' +  IntToHex(LoWord(DI.Serial), 4));
          end; {if}
          SubItems.Add(Drv);
        except
        //  Items.Delete(NewItem.Index);
        end; {try}
      end; {with}
    end; {if}
  end; {for}
  lvDrives.Items.EndUpdate;
end; {AddDrives}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.UpdateFileListView(FileMask : string; Attribute : DWORD);
var
  FileCount : integer;                                                    // No. files found.
  DirCount  : integer;                                                    // No. directories found.

  {................................................................................................}

  function  DefaultFileType(Filename : string; FileAttributes : DWORD) : string;
  begin
    if (FileAttributes and FILE_ATTRIBUTE_DIRECTORY > 0) then
      Result  := 'File Folder'
    else begin
        Result  := UpperCase(ExtractFileExt(Filename));
      if (Result = '.') then
        Result  := ''
      else
        Result  := Copy(Result, 2, pred(Length(Result))) + ' File';
    end; {if}
  end; {DefaultFileType}

  {................................................................................................}

  function AttributeString(Attribute  : DWORD) : string;                  // Generate a string based on
  begin                                                                   // the given attribute word.
    Result  := '';
    if (FILE_ATTRIBUTE_READONLY and Attribute > 0) then
      Result  := 'R';
    if (FILE_ATTRIBUTE_HIDDEN   and Attribute > 0) then
      Result  := Result + 'H';
    if (FILE_ATTRIBUTE_SYSTEM   and Attribute > 0) then
      Result  := Result + 'S';
    if (FILE_ATTRIBUTE_ARCHIVE  and Attribute > 0) then
      Result  := Result + 'A';
    if (FILE_ATTRIBUTE_NORMAL   and Attribute > 0) then
      Result  := Result + 'N';
    if (Result = '') then
      Result  := '.';
  end; {AttributeString}

  {................................................................................................}

  procedure AddFileToList(SearchRec : TSearchRec);                        // Add a file to the list.
  var
    ShInfo    : TSHFileInfo;
    Filename  : string;
  begin
    Filename  := ExtractFilePath(fFilename) + SearchRec.Name;
    SHGetFileInfo(PChar(Filename), 0, ShInfo, SizeOf(ShInfo), SHGFI_TYPENAME or SHGFI_SYSICONINDEX or SHGFI_DISPLAYNAME);
    with lvFiles.Items.Add, SearchRec.FindData do begin
      if (ShInfo.szDisplayName <> '') then                                // If it's got a name then use it ...
        Caption := ShInfo.szDisplayName
      else                                                                // otherwise use this one instead.
        Caption := SearchRec.Name;
      if (SearchRec.Name = '..') then begin                               // If it's the parent root directory then
        Caption     := ExtractFilename(ExpandFileName(FileName));         // get its name from the filename ...
        if (Caption = '') then                                            // might be nothing left at the root, so give
          Caption := '..';                                                // it the dot dot ...
        Caption     := Format('[%s]', [Caption]);                         // Add the square brackets and give it one of our
        if (lvFiles.SmallImages <> nil) then                              // home-made icons.
          ImageIndex  := fRootIconIndex
      end else if (lvFiles.SmallImages <> nil) then                       // Otherwise use an icon that Bill made !
          ImageIndex  := ShInfo.iIcon;
      if ((dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) > 0) then       // Everyone assumes that directories take no space ...
        SubItems.Add('')                                                  // and so do I ... so don't give it a size !
      else                                                                // Otherwise ... give its size in bytes.
        SubItems.Add(FormatFloat('#0,', nFileSizeHigh * MAXDWORD + nFileSizeLow));
      if (SHInfo.szTypeName = '') then                                    // Add on its type name ... look out for
        SubItems.Add(DefaultFileType(Filename, dwFileAttributes))         // the shell not recognising the file type ...
      else                                                                // so we have to do it ourselves.
        SubItems.Add(ShInfo.szTypeName);
      SubItems.Add(FileTimeToDateTimeStr(ftCreationTime  ));              // Now add on the file time stamp stuff.
      SubItems.Add(FileTimeToDateTimeStr(ftLastWriteTime ));              // Be aware that 95 and 98 always has
      SubItems.Add(FileTimeToDateTimeStr(ftLastAccessTime));              // zero accessed time.
      if ((dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) > 0) then begin
        if (SearchRec.Name = '..') then begin
          Filename  := Copy(Filename, 1, Length(Filename) - 2);
          SubItems.Add(AttributeString(GetFileAttributes(PChar(Filename))));
          SubItems.Add(Filename);
        end else begin
          SubItems.Add(AttributeString(dwFileAttributes));                // These are invisible but needed - long filename.
          SubItems.Add(ExpandFileName(Filename));                         // Now a flag character to say whether it's
          inc(DirCount);                                                  // a directory or a file ... and while we're
        end; {if}                                                         // at it, update our counts.
        SubItems.Add('D');
      end else begin
        SubItems.Add(AttributeString(dwFileAttributes));                  // Now its attributes.
        SubItems.Add(Filename);                                           // These are invisible but needed - long filename.
        inc(FileCount);
        SubItems.Add('F');
      end; {if}
    end; {with}
  end; {AddFileToList}
  
  {................................................................................................}

var
  lp1           : integer;
  SearchRec     : TSearchRec;
  HasDriveRoot  : boolean;
begin
  if SetCurrentDir(ExtractFilePath(fFilename)) then begin
    FileMask  := '*.*';
    FileCount := 0;
    DirCount  := 0;

    lvFiles.Items.BeginUpdate;
    lvFiles.Items.Clear;

    HasDriveRoot  := (ExtractFileDir(fFilename) = ExtractFilePath(fFilename));
    if HasDriveRoot then begin
      with lvFiles.Items.Add do begin
        Caption := '[Drives]';
        if (lvFiles.SmallImages <> nil) then
          ImageIndex  := fRootIconIndex;
        for lp1 := cColSize to cColFullName do
          SubItems.Add('');
        SubItems.Add('D');                                                                                                  // and so do I ... so don't give it a size !
      end; {with}
    end; {if}

    if (FindFirst(FileMask, faAnyFile, SearchRec) = 0) then begin
      try
        Screen.Cursor := crHourGlass;
        repeat
          if (SearchRec.Name <> '.') and not (HasDriveRoot and (SearchRec.Name = '..')) then begin
            if (SearchRec.Attr and FILE_ATTRIBUTE_DIRECTORY = 0) then begin // Don't come in here with these boys ...
              with SearchRec.FindData do begin
                if (((Attribute and dwFileAttributes and FILE_ATTRIBUTE_READONLY > 0))  or
                    ((Attribute and dwFileAttributes and FILE_ATTRIBUTE_HIDDEN   > 0))  or
                    ((Attribute and dwFileAttributes and FILE_ATTRIBUTE_SYSTEM   > 0))  or
                    ((Attribute and dwFileAttributes and FILE_ATTRIBUTE_ARCHIVE  > 0))  or
                    ((Attribute and dwFileAttributes and FILE_ATTRIBUTE_NORMAL   > 0))) or
                    (dwFileAttributes = 0) then
                  AddFileToList(SearchRec);                                 // OK ... put me on the list.
              end; {with}
            end else if not fFoldersHidden then
              AddFileToList(SearchRec);
          end; {if}
        until (FindNext(SearchRec) <> 0);                                   // Keep on looking.
      finally
        FindClose(SearchRec);                                               // Close and put the cursor back.
        Screen.Cursor := crDefault;
      end; {try}
    end; {if}
                                                                            // Release the list view.
    lvFiles.Items.EndUpdate;

    StatusBar.Panels[0].Text  := 'Files: '   + IntToStr(FileCount);         // Tell us how many we got.
    StatusBar.Panels[1].Text  := 'Folders: ' + IntToStr(DirCount );

    lvDrives.Visible          := false;
    lvFiles.Visible           := true;

  end; {if}
end; {UpdateFileListView}

{--------------------------------------------------------------------------------------------------}
{                                         Menu Events                                              }
{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuAboutClick(Sender : TObject);
// What else ? - it shows the about box !
begin
  with TfrmAbout.Create(nil) do begin
    try
      if fIsPropertyEditor then
        Caption  := 'Key Maker Package'
      else
        Caption  := 'File Browser';
      ShowModal;
    finally
      Free;
    end; {try}
  end; {with}
end; {mnuAboutClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuAttributesClick(Sender : TObject);
// Revise the file attributes filter.
begin
  TMenuItem(Sender).Checked := not TMenuItem(Sender).Checked;             // Toggle me !
  SetFileAttributesFilter;                                                // Set up the attributes filter.
  UpdateFileListView(fFileMask, fFileAttributes);                         // Go and get the boys in the list.
end; {mnuAttributesClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuBookmarkClick(Sender : TObject);
begin
  if (lvFiles.Selected <> nil) then begin
    if (TWinControl(Sender).Tag = cAddBookmark) then
      fajRecentKeys.UpdateRecentKeys(lvFiles.Selected.SubItems.Strings[pred(cColFullName)], tAdd)
    else
      fajRecentKeys.UpdateRecentKeys(lvFiles.Selected.SubItems.Strings[pred(cColFullName)], tDelete);
    SetBookmarkControls(TWinControl(Sender).Tag);
  end; {if}
end; {mnuBookmarkClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuColumnsClick(Sender : TObject);
// Hide or reveal chosen column - restores previously used column width.
begin
  with TMenuItem(Sender) do begin
    Checked := not Checked;                                               // Toggle me too !
    fColumnSettings[tag].Visible  := Checked;                             // Note that the menus use the tags.
    if Checked then                                                       // If visible now then restore
      lvFiles.Columns[tag].Width  := fColumnSettings[tag].DefWidth        // the last known width.
    else                                                                  // Otherwise conceal this column
      lvFiles.Columns[tag].Width  := 0;                                   // by setting its width to zero.
  end; {with}                                                             // Hey ... don't forget that the new
end; {mnuColumnsClick}                                                    // window proc will have done its stuff too.

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuCopyPathClick(Sender : TObject);
var
  Item  : TListItem;
  Path  : string;
begin
  Path  := '';
  Item  := lvFiles.Selected;
  repeat
    Path  := Path + Item.SubItems[pred(cColFullName)] + #13;
    Item  := lvFiles.GetNextItem(Item, sdAll, [isSelected]);
  until (Item = nil);
  Path  := Copy(Path, 1, pred(Length(Path)));
  Clipboard.SetTextBuf(PChar(Path));
end; {mnuCopyPathClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuEditFileParamsClick(Sender : TObject);
var
  NewParameters : TWin32FindData;
  Filename      : string;
  OldAttributes : DWORD;
  Item          : TListItem;
begin
  with TfrmEditFileParams.Create(nil) do begin
    FillMemory(@NewParameters, SizeOf(NewParameters), 0);
    try
      Item  := lvFiles.Selected;
      while (Item <> nil) do begin
        Filename  := ExcludeTrailingBackSlash(Item.SubItems[pred(cColFullName)]);

        if GetFileAttributesEx(PChar(Filename), GetFileExInfoStandard, @Win32FindData) then begin
          with Win32FindData do begin
            OldAttributes := dwFileAttributes;
            SetFileAttributes(PChar(Filename), 0);
            FileHandle    := CreateFile(PChar(Filename),
                                        GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE,
                                        nil, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH or FILE_FLAG_BACKUP_SEMANTICS, 0);
            if (ModalResult = mrNone) then begin
              if (Item.SubItems[pred(cColDirFile)] = 'D') then
                Caption := 'Edit Dir Params - '  + ExtractFileName(Filename)
              else
                Caption := 'Edit File Params - ' + ExtractFileName(Filename);
              ShowModal;
              Update;
              NewParameters.dwFileAttributes  := dwFileAttributes;
              NewParameters.ftCreationTime    := ftCreationTime;
              NewParameters.ftLastWriteTime   := ftLastWriteTime;
              NewParameters.ftLastAccessTime  := ftLastAccessTime;
            end; {if}
          end; {with}
          with NewParameters do begin
            if (ModalResult = mrOK) then begin
              if (FileHandle <> INVALID_HANDLE_VALUE) then begin
                if (ftCreationTime.dwLowDateTime   <> 0) or (ftCreationTime.dwHighDateTime   <> 0) then
                  SetFileTime(FileHandle, @ftCreationTime, nil, nil  );
                if (ftLastWriteTime.dwLowDateTime  <> 0) or (ftLastWriteTime.dwHighDateTime  <> 0) then
                  SetFileTime(FileHandle, nil, nil, @ftLastWriteTime );
                if (ftLastAccessTime.dwLowDateTime <> 0) or (ftLastAccessTime.dwHighDateTime <> 0) then
                  SetFileTime(FileHandle, nil, @ftLastAccessTime, nil);
                CloseHandle(FileHandle);
                FileHandle  := INVALID_HANDLE_VALUE;
              end; {if}
              if SaveAttributes then
                SetFileAttributes(PChar(Filename), dwFileAttributes)
              else
                SetFileAttributes(PChar(Filename), OldAttributes);
            end else
              SetFileAttributes(PChar(Filename), OldAttributes);
            if (FileHandle <> INVALID_HANDLE_VALUE) then
              CloseHandle(FileHandle);
          end; {with}
        end; {if}
        Item  := lvFiles.GetNextItem(Item, sdAll, [isSelected]);
      end; {while}
    finally
      Free;
    end; {try}
  end; {with}
  RefreshlvFiles;
end; {mnuEditFileTimeClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuExitClick(Sender : TObject);
// Looks like we're leaving !
begin
  if fIsPropertyEditor then
    ModalResult := mrCancel
  else
    Close;
end; {mnuExitClick}

{--------------------------------------------------------------------------------------------------}

function TfrmFileBrowser.FileFindAndSelect(ListView : TListView; FileCaption : string) : boolean;
var
  ListItem  : TListItem;
begin
  ListItem  := ListView.FindCaption(0, FileCaption, false, true, false);
  Result    := (ListItem <> nil);
  if Result then
    ListView.ItemFocused   := ListItem;
  HighlightFocusedItem(ListView);
end; {FileFindAndSelect}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.RefreshlvFiles;
var
  Caption : string;
begin
  if (lvFiles.Selected <> nil) then
    Caption := lvFiles.Selected.Caption
  else
    Caption := '';
  UpdateFileListView(fFileMask, fFileAttributes);                         // Go and get the same old stuff again.
  FileFindAndSelect(lvFiles, Caption);
end; {RefreshlvFiles}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.RefreshlvDrives;
var
  Caption : string;
begin
  if (lvDrives.Selected <> nil) then
    Caption := lvDrives.Selected.Caption
  else
    Caption := '';
  AddDrives;
  FileFindAndSelect(lvDrives, Caption);
end; {RefreshlvDrives}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuRefreshClick(Sender : TObject);
// Guess what ? - it refreshes the current list view ... just in case something changed !
begin
  if lvDrives.Visible then
    RefreshlvDrives
  else
    RefreshlvFiles;
end; {mnuRefreshClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuShowFoldersClick(Sender : TObject);
// Hide or reveal folders.
begin
  TMenuItem(Sender).Checked := not TMenuItem(Sender).Checked;             // Toggle me.
  fFoldersHidden            := not TMenuItem(Sender).Checked;             // Decide whether we get the folders.
  if lvFiles.Visible then
    UpdateFileListView(fFileMask, fFileAttributes);                       // ... and just do it.
end; {mnuShowFoldersClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuUpdateClick(Sender : TObject);
// Closing time but tell the property manager that it's OK to update the new value.
begin
  if fIsPropertyEditor then
    ModalResult := mrOK
  else
    Close;
end; {mnuUpdateClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuViewTypeClick(Sender : TObject);
// Select view preference ... large icons, small icons, list or details.
begin
  if (Sender is TMenuItem) then begin
    with Sender as TMenuItem do begin
      case Tag of
        0 : btnLargeIcons.Down    := true;
        1 : btnSmallIcons.Down    := true;
        2 : btnList.Down          := true;
        3 : btnDetails.Down       := true;
      end; {case}
      Checked             := true;                                            // Toggle me.
      lvDrives.ViewStyle  := TViewStyle(Tag);
      lvFiles.ViewStyle   := TViewStyle(Tag);                                 // Note that the menu tags are used.
    end; {with}
  end else if (Sender is TSpeedButton) then begin
    with Sender as TSpeedButton do begin
      case Tag of
        0 : mnuLargeIcons.Checked := true;
        1 : mnuSmallIcons.Checked := true;
        2 : mnuList.Checked       := true;
        3 : mnuDetails.Checked    := true;
      end; {case}
      lvDrives.ViewStyle  := TViewStyle(Tag);
      lvFiles.ViewStyle   := TViewStyle(Tag);                                 // Note that the button tags are used.
    end; {with}
  end; {if}
  if (lvFiles.ViewStyle = vsReport) then
    SetColumnBitmap(lvFiles.Column[fColumnToSort]);
end; {mnuViewTypeClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.mnuXPMenuClick(Sender : TObject);
// Set XP menu styles.
begin
  with Sender as TMenuItem do begin
    Checked               := not Checked;                                 // Toggle me.
    fajAlterMenu.Enabled  := Checked;                                     // Switch on the fancy stuff ...
  end; {with}                                                             // or maybe switch us off.
end; {mnuXPMenuClick}

{--------------------------------------------------------------------------------------------------}
{                                         Other Events                                             }
{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.SetColumnBitmap(Column : TListColumn);          // Gosh ... all this just to put
var                                                                       // an up/down arrow in the column
  HeaderHandle  : THandle;                                                // header ... lot of work !
  HdItem        : THdItem;
  lp1           : Integer;
begin
  HeaderHandle  := ListView_GetHeader(lvFiles.Handle);
  if (HeaderHandle <> 0) then begin
    for lp1 := 0 to pred(lvFiles.Columns.Count) do begin
      HdItem.Mask := HDI_FORMAT;
      Header_GetItem(HeaderHandle, lp1, HdItem);
      HdItem.Mask := HDI_BITMAP or HDI_FORMAT;
      if (lp1 = Column.Index) then begin
        HdItem.hbm  := fSortArrows[ord(fSortAscending)].Handle;
        HdItem.fmt  := HdItem.fmt or HDF_BITMAP;
        if (lvFiles.Columns[lp1].Alignment = taLeftJustify) then
          HdItem.fmt  := HdItem.fmt or HDF_BITMAP_ON_RIGHT;
      end else
        HdItem.fmt  := HdItem.fmt and not HDF_BITMAP;
      Header_SetItem(HeaderHandle, lp1, HdItem);
    end; {for}
  end; {if}
end; {SetColumnBitmap}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvFilesColumnClick(Sender : TObject; Column : TListColumn);
begin
  if (fColumnToSort = Column.Index) then                                  // Set the sort direction.
    fSortAscending  := not fSortAscending
  else begin
    fColumnToSort   := Column.Index;
    fSortAscending  := true;
  end;
  (Sender as TCustomListView).AlphaSort;                                 // Execute the sort and then
  SetColumnBitmap(Column);                                               // show us which column.
end; {lvFilesColumnClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvFilesCompare(Sender : TObject; Item1, Item2 : TListItem; Data : integer; var Compare : integer);

  {................................................................................................}

  function NumberStringToInt64(NumStr : string) : Int64;                  // Convert a numeric string to an Int64.
  var                                                                     // Better than StrToIntDef ...
    lp1 : integer;                                                        // This one deals with anthying that's not
  begin                                                                   // numeric by ignoring it.
    Result  := 0;
    for lp1 := 1 to Length(NumStr) do
      if NumStr[lp1] in ['0'..'9'] then
        Result  := Result * 10 + ord(NumStr[lp1]) - 48;
  end; {NumberStringToDWORD}

  {................................................................................................}

  function CompareFileSize(DirFile1, Filename1, SizeStr1, DirFile2, Filename2, SizeStr2 : string) : integer;
  var
    Num1  : Int64;                                                        // Compare file sizes.
    Num2  : Int64;
  begin
    if (DirFile1 = 'D') and (DirFile2 = 'D') then                         // Directories have no size ...
      Result  := CompareText(Filename1, Filename2)                        // so sort them by name.
    else begin
      Num1  := NumberStringToInt64(SizeStr1);                             // Get the sizes and compare them.
      Num2  := NumberStringToInt64(SizeStr2);
      if (Num1 > Num2) then
        Result  := 1
      else if (Num1 < Num2) then
        Result  := -1
      else                                                                // Same size so sort them by name.
        Result  := CompareText(Filename1, Filename2);
    end; {if}
  end; {CompareFileSize}

  {................................................................................................}

  function ReconstructDate(DateStr : string) : TDateTime;
  var                                                                     // This one was hard.
    TimePos     : integer;                                                // Parse out the date and convert
    h, m, s, n  : WORD;                                                   // it back to TDateTime.
  begin                                                                   // Then parse out the time and add
    TimePos := Pos(' ', DateStr);                                         // that to our TDateTime ... and then
    if (TimePos = 0) then                                                 // we can do the comparison.
      Result  := StrToDate(DateStr)                                       
    else begin
      Result  := StrToDate(Copy(DateStr, 1, pred(TimePos)));
      DateStr := Copy(DateStr, succ(TimePos), Length(DateStr)-TimePos);
      h       := StrToInt(Copy(DateStr,  1, 2));
      m       := StrToInt(Copy(DateStr,  4, 2));
      s       := StrToInt(Copy(DateStr,  7, 2));
      n       := StrToInt(Copy(DateStr, 10, 3));
      Result  := Result + EncodeTime(h, m, s, n);
    end; {if}
  end; {ReconstructDate}

  {................................................................................................}

  function CompareDates(DateStr1, Filename1, DateStr2, Filename2 : string) : integer;
  var
    Date1, Date2  : TDateTime;                                            // Compare the dates.
  begin
    Date1 := ReconstructDate(DateStr1);                                   // Well get the dates first.
    Date2 := ReconstructDate(DateStr2);
    if (Date1 > Date2) then
      Result  := 1
    else if (Date1 < Date2) then
      Result  := -1
    else                                                                  // Same dates so sort them by name.
      Result  := CompareText(Filename1, Filename2);
  end; {CompareDates}

  {................................................................................................}

var
  DirFile1, DirFile2  : string;
begin
  DirFile1  := Item1.SubItems.Strings[pred(cColDirFile)];                 // Get the file type ... D or F.
  DirFile2  := Item2.SubItems.Strings[pred(cColDirFile)];

  if (Pos('[', Item1.Caption) = 1) or (Pos('[', Item2.Caption) = 1) then begin
                                                                          // Must keep the parent root at the
    if (Pos('[', Item1.Caption) = 1) then                                 // top of the list at all costs.
      Compare := -1                                                       // The '[' uniquely defines this
    else                                                                  // entry.
      Compare := 1;

  end else begin

    if (DirFile1 = 'D') and (DirFile2 = 'F') then                         // If one's a file and the other's a
                                                                          // directory then it's a no brainer ...
      Compare  := -1                                                      // directories come first or last
                                                                          // depending on the sort direction.
    else if (DirFile1 = 'F') and (DirFile2 = 'D') then

      Compare  := 1

    else begin

      case fColumnToSort of                                               // Choose our column ...
        cColName        : Compare := CompareText(DirFile1 + Item1.Caption,
                                                 DirFile2 + Item2.Caption);

        cColSize        : Compare := CompareFileSize(DirFile1, Item1.Caption, Item1.SubItems.Strings[pred(fColumnToSort)],
                                                     DirFile2, Item2.Caption, Item2.SubItems.Strings[pred(fColumnToSort)]);

        cColType        : Compare := CompareText(Item1.SubItems.Strings[pred(fColumnToSort)] + Item1.Caption,
                                                 Item2.SubItems.Strings[pred(fColumnToSort)] + Item2.Caption);

        cColCreated,
        cColModified,
        cColAccessed    : Compare := CompareDates(Item1.SubItems.Strings[pred(fColumnToSort)], Item1.Caption,
                                                  Item2.SubItems.Strings[pred(fColumnToSort)], Item2.Caption);

        cColAttributes  : Compare := CompareText(DirFile1 + Item1.SubItems.Strings[pred(fColumnToSort)] + Item1.Caption,
                                                 DirFile2 + Item2.SubItems.Strings[pred(fColumnToSort)] + Item2.Caption);

        cColFullName    : Compare := CompareText(Item1.SubItems.Strings[pred(fColumnToSort)],
                                                 Item2.SubItems.Strings[pred(fColumnToSort)]);

      end; {case}

    end; {if}

    if not fSortAscending then                                            // Switch our return value depending
      Compare := Compare * -1;                                            // on the sort direction.

  end; {if}

end; {lvFilesCompare}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.FileSelected(Item : TListItem);
begin
  if (Item <> nil) then begin
    if (Item.SubItems[pred(cColDirFile)] = 'D') then begin
      fFilename := Item.SubItems[pred(cColFullName)];
      if (fFilename <> '') then begin
        if (fFilename[Length(fFilename)] = '\') then
          fFilename := ExtractFilePath(ExcludeTrailingBackslash(fFilename))
        else
          fFilename := fFilename + '\';
      end; {if}
      edFilename.Text := fFilename;
      if (fFilename = '') then begin
        StatusBar.Panels[0].Text  := '';
        StatusBar.Panels[1].Text  := '';
        lvDrives.Visible          := true;
        lvFiles.Visible           := false;
        HighlightFocusedItem(lvDrives);
      end else begin
        lvDrives.Visible          := false;
        lvFiles.Visible           := true;
        UpdateFileListView(fFileMask, fFileAttributes);
        HighlightFocusedItem(lvFiles);
      end; {if}
      SetBookmarkControls(mnuPopBookmark.Tag);
    end else                                                              // or a file ... so use its association.
      ShellExecute(Application.Handle, 'Open', PChar(Item.SubItems[pred(cColFullName)]), nil, nil, SW_NORMAL);
  end; {if}
end; {FileSelected}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvFilesDblClick(Sender : TObject);
begin
  FileSelected(lvFiles.Selected);                                         // A double click ... so do it !
end; {lvFilesDblClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvFilesKeyDown(Sender : TObject; var Key : Word; Shift : TShiftState);
begin
  with lvFiles do begin
    case Key of
      VK_LEFT   : if (Pos('[', Items[0].Caption) = 1) and (lvFiles.ViewStyle = vsReport) then begin
                    FileSelected(Items[0]);
                    Key := 0;
                  end; {if}
      VK_RIGHT  : if  (Selected <> nil) and (lvFiles.ViewStyle = vsReport) then begin
                    if (Pos('[', Selected.Caption) = 0) and
                         (Selected.SubItems[pred(cColDirFile)] = 'D') then
                      FileSelected(Selected);
                      Key := 0;
                  end; {if}
      VK_RETURN : FileSelected(Selected);                                 // Return pressed ... so do that too !
    end; {case}
  end; {with}
end; {lvFilesKeyDown}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvFilesSelectItem(Sender : TObject; Item : TListItem; Selected : boolean);
begin
  if Selected then begin
    lvFiles.PopupMenu := lvFilesPopupMenu;
    fFilename         := Item.SubItems.Strings[pred(cColFullName)];
    edFilename.Text   := fFilename;
    SetBookmarkControls(mnuPopBookmark.Tag);
  end else
    lvFiles.PopupMenu := nil;
end; {lvFilesSelectItem}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.RecentKeyClick(Keyname : string);
// Recover a previous bookmark - directory or file.
begin
  Update;
  Filename  := Keyname;
  if (Keyname[Length(KeyName)] = '\') then begin
    KeyName := ExtractFilePath(ExcludeTrailingBackslash(KeyName));
    KeyName := '[' + ExtractFileName(ExcludeTrailingBackslash(KeyName)) + ']';
    if (Keyname = '[]') then
      Keyname := '[..]';
  end; {if}
  if FileFindAndSelect(lvFiles, ExtractFilename(Keyname)) then
    fajRecentKeys.UpdateRecentKeys(Keyname, tAdd)
  else
    fajRecentKeys.UpdateRecentKeys(Keyname, tDelete);
  SetBookmarkControls(mnuPopBookmark.Tag);
end; {RecentFileClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvDrivesCompare(Sender : TObject; Item1, Item2 : TListItem; Data : integer; var Compare : integer);
begin
  Compare := CompareText(Item1.SubItems.Strings[cColDriveName], Item2.SubItems.Strings[cColDriveName]);
end; {lvDrivesCompare}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvDrivesDblClick(Sender : TObject);
begin
  if (lvDrives.Selected <> nil) then begin
    fFilename := lvDrives.Selected.SubItems.Strings[cColDriveName];
    UpdateFileListView(fFileMask, fFileAttributes);
    HighlightFocusedItem(lvFiles);
  end; {if}
end; {lvDrivesDblClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.SetBookmarkControls(Tag : integer);
var
  Glyph           : TBitmap;
  EnableBookmarks : boolean;
begin
  EnableBookmarks := lvFiles.Visible and (lvFiles.Selected <> nil);
  if EnableBookmarks then
    EnableBookmarks := (lvFiles.Selected.SubItems.Strings[pred(cColFullName)] <> '');
  if EnableBookmarks then begin
    btnBookmark.Enabled     := true;
    mnuPopBookmark.Enabled  := true;
    if fajRecentKeys.InList(lvFiles.Selected.SubItems.Strings[pred(cColFullName)]) then begin
      if (Tag = cAddBookmark) then begin
        Glyph := TBitmap.Create;
        imButtons.GetBitmap(1, Glyph);
        btnBookmark.Glyph.Assign(Glyph);
        Glyph.Free;
        btnBookmark.Hint          := 'Remove bookmark';
        btnBookmark.Tag           := cRemoveBookmark;
        mnuPopBookmark.ImageIndex := cRemoveBookmark;
        mnuPopBookmark.Caption    := '&Remove bookmark';
        mnuPopBookmark.Tag        := cRemoveBookmark;
      end; {if}
    end else begin
      if (Tag = cRemoveBookmark) then begin
        Glyph := TBitmap.Create;
        imButtons.GetBitmap(0, Glyph);
        btnBookmark.Glyph.Assign(Glyph);
        Glyph.Free;
        btnBookmark.Hint          := 'Add bookmark';
        btnBookmark.Tag           := cAddBookmark;
        mnuPopBookmark.ImageIndex := cAddBookmark;
        mnuPopBookmark.Caption    := '&Add bookmark';
        mnuPopBookmark.Tag        := cAddBookmark;
      end; {if}
    end; {if}
  end else begin
    btnBookmark.Enabled     := false;
    mnuPopBookmark.Enabled  := false;
  end; {if}
end; {SetBookmarkControls}

{--------------------------------------------------------------------------------------------------}

procedure TfrmFileBrowser.lvDrivesKeyDown(Sender : TObject; var Key : Word; Shift : TShiftState);
begin
  with lvDrives do begin
    case Key of
      VK_LEFT   : Key := 0;
      VK_RIGHT  : if (lvDrives.ViewStyle = vsReport) then begin
                    fFilename := lvDrives.Selected.SubItems.Strings[cColDriveName];
                    UpdateFileListView(fFileMask, fFileAttributes);
                    HighlightFocusedItem(lvFiles);
                    Key := 0;
                  end; {if}
      VK_RETURN : begin
                    fFilename := lvDrives.Selected.SubItems.Strings[cColDriveName];
                    UpdateFileListView(fFileMask, fFileAttributes);
                    HighlightFocusedItem(lvFiles);
                  end;
    end; {case}
  end; {with}
end; {lvDrivesKeyDown}

{--------------------------------------------------------------------------------------------------}
{ajFileBrowser}
end.
