          ///////////////////////////////////////////////////////////////////////
          //                                                                   //
          //       SoftSpot Software Component Library                         //
          //       Key Maker software component                                //
          //                                                                   //
          //       Copyright (c) 1996 - 2002 SoftSpot Software Ltd.            //
          //       ALL RIGHTS RESERVED                                         //
          //                                                                   //
          //   The entire contents of this file is protected by U.S. and       //
          //   International Copyright Laws. Unauthorized reproduction,        //
          //   reverse-engineering, and distribution of all or any portion of  //
          //   the code contained in this file is strictly prohibited and may  //
          //   result in severe civil and criminal penalties and will be       //
          //   prosecuted to the maximum extent possible under the law.        //
          //                                                                   //
          //   RESTRICTIONS                                                    //
          //                                                                   //
          //   THE SOURCE CODE CONTAINED WITHIN THIS FILE AND ALL RELATED      //
          //   FILES OR ANY PORTION OF ITS CONTENTS SHALL AT NO TIME BE        //
          //   COPIED, TRANSFERRED, SOLD, DISTRIBUTED, OR OTHERWISE MADE       //
          //   AVAILABLE TO OTHER INDIVIDUALS WITHOUT EXPRESS WRITTEN CONSENT  //
          //   AND PERMISSION FROM SOFTSPOT SOFTWARE LTD.                      //
          //                                                                   //
          //   CONSULT THE END USER LICENSE AGREEMENT FOR INFORMATION ON       //
          //   ADDITIONAL RESTRICTIONS.                                        //
          //                                                                   //
          ///////////////////////////////////////////////////////////////////////

unit ajRegistryBrowser;

{                             -=< Key Maker Registry Browser >=-
{
{ Copyright SoftSpot Software 2002 - All Rights Reserved
{
{ Author        : Andrew J Jameson
{
{ Web site      : www.softspotsoftware.com
{ e-mail        : contact@softspotsoftware.com
{
{ Creation Date : 01 March 2002
{
{ Version       : 1.00
{
{ Description   : Registry property editor for Key Maker.                                          }

interface

uses
  Forms, Windows, Controls, ExtCtrls, ComCtrls, StdCtrls, Classes, Buttons,
  SysUtils, ImgList, Registry, Menus, ajRecentKeys, ajKeyMakerAbout, ajAlterMenu;

type
  TajBrowserException = class(Exception);

type
  TajRegPathManager = class(TObject)
  private
    fRootHKEY     : HKEY;
    fRootHKEYText : string;
    fFullPath     : string;
    fSubPath      : string;
  protected
    procedure SetfFullPath  (Value : string);
  public
    property  RootHKEY      : HKEY      read fRootHKEY;
    property  RootHKEYText  : string    read fRootHKEYText;
    property  FullPath      : string    read fFullPath  write SetfFullPath;
    property  SubPath       : string    read fSubPath;
  end;

type
  TfrmRegBrowser = class(TForm)
    Bevel1          : TBevel;
    Bevel2          : TBevel;
    btnAddGuid      : TSpeedButton;
    btnBookmark     : TSpeedButton;
    btnCancel       : TSpeedButton;
    btnRefresh      : TSpeedButton;
    btnUpdate       : TSpeedButton;
    edRegPath       : TEdit;
    imButtons       : TImageList;
    imTreeView      : TImageList;
    lvPopupMenu     : TPopupMenu;
    lvRegistry      : TListView;
    MainMenu        : TMainMenu;
    mnuAbout        : TMenuItem;
    mnuAddGUID      : TMenuItem;
    mnuBinary       : TMenuItem;
    mnuBookmarks    : TMenuItem;
    mnuDeleteValue  : TMenuItem;
    mnuDWORD        : TMenuItem;
    mnuEditValue    : TMenuItem;
    mnuExit         : TMenuItem;
    mnuNewValue     : TMenuItem;
    mnuOptions      : TMenuItem;
    mnuPopAddKey    : TMenuItem;
    mnuPopBookmark  : TMenuItem;
    mnuPopDeleteKey : TMenuItem;
    mnuRefresh      : TMenuItem;
    mnuRegistry     : TMenuItem;
    mnuString       : TMenuItem;
    mnuUpdate       : TMenuItem;
    mnuView         : TMenuItem;
    mnuXPMenu       : TMenuItem;
    N1              : TMenuItem;
    N2              : TMenuItem;
    N4              : TMenuItem;
    pnBack          : TPanel;
    pnRegPath       : TPanel;
    pnToolbar       : TPanel;
    Splitter        : TSplitter;
    StatusBar       : TStatusBar;
    tvPopupMenu     : TPopupMenu;
    tvRegistry      : TTreeView;
    procedure FormCreate            (Sender : TObject);
    procedure FormShow              (Sender : TObject);
    procedure FormClose             (Sender : TObject; var Action : TCloseAction);
    procedure FormDestroy           (Sender : TObject);
    procedure mnuAboutClick         (Sender : TObject);
    procedure mnuAddGUIDClick       (Sender : TObject);
    procedure mnuAddKeyClick        (Sender : TObject);
    procedure mnuBookmarkClick      (Sender : TObject);
    procedure mnuDeleteKeyClick     (Sender : TObject);
    procedure mnuDeleteValueClick   (Sender : TObject);    
    procedure mnuExitClick          (Sender : TObject);
    procedure mnuRefreshClick       (Sender : TObject);
    procedure mnuUpdateClick        (Sender : TObject);
    procedure mnuXPMenuClick        (Sender : TObject);
    procedure tvRegistryChange      (Sender : TObject; Node : TTreeNode);
    procedure tvRegistryExpanding   (Sender : TObject; Node : TTreeNode; var AllowExpansion : boolean);
    procedure tvRegistryMouseDown   (Sender : TObject; Button : TMouseButton; Shift : TShiftState; X, Y : integer);
    procedure lvPopupMenuPopup      (Sender : TObject);
  private
    { Private declarations }
    fajAlterMenu      : TajAlterMenu;
    fajRecentKeys     : TajRecentKeys;
    fRegPathManager   : TajRegPathManager;
    fIsPropertyEditor : boolean;
  protected
    procedure SetfIsPropertyEditor  (Value : boolean);
    procedure AddListNames          (SubPath : string; List : TStringList);
    procedure AddTreeNodes          (Node : TTreeNode; List : TStrings);
    procedure GetRegKeyValueNames   (SubPath : string);
    procedure GetRegKeyNames        (SubPath : string; Node : TTreeNode);
    procedure RecentKeyClick        (KeyName : string);
    function  GetRegistryPath       : string;
    procedure SetRegistryPath       (Value : string);
    procedure SetBookmarkControls   (Tag : integer);
    procedure DeleteBranchRoot      (HKEYText : string);
    procedure AddKey                (FullPath : string);
    procedure DeleteKey             (FullPath : string);
    procedure DeleteValue           (KeyValue : string);    
  public
    { Public declarations }
    function  GoToBookmark          (FullPath : string) : boolean;
    property  IsPropertyEditor  : boolean   read fIsPropertyEditor  write SetfIsPropertyEditor;
    property  RegistryPath      : string    read GetRegistryPath    write SetRegistryPath;
  end;

var
  frmRegBrowser : TfrmRegBrowser;

implementation

{$R *.DFM}

{$IFDEF VER140}   // Delphi6
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}

uses
  Graphics, CommCtrl, Dialogs, ComObj, ActiveX, ajRegistry;

const

  {................................................................................................}

  cNumRegValType  = 11; // Number of type supported.
  cRegValTypeLookUp : array[0..pred(cNumRegValType)] of record
    Description : string;
    Value       : DWORD;
  end = ((Description : 'REG_NONE';                       Value : REG_NONE),
         (Description : 'REG_SZ';                         Value : REG_SZ),
         (Description : 'REG_EXPAND_SZ';                  Value : REG_EXPAND_SZ),
         (Description : 'REG_BINARY';                     Value : REG_BINARY),
         (Description : 'REG_DWORD';                      Value : REG_DWORD),
         (Description : 'REG_DWORD_BIG_ENDIAN';           Value : REG_DWORD_BIG_ENDIAN),
         (Description : 'REG_LINK';                       Value : REG_LINK),
         (Description : 'REG_MULTI_SZ';                   Value : REG_MULTI_SZ),
         (Description : 'REG_RESOURCE_LIST';              Value : REG_RESOURCE_LIST),
         (Description : 'REG_FULL_RESOURCE_DESCRIPTOR';   Value : REG_FULL_RESOURCE_DESCRIPTOR),
         (Description : 'REG_RESOURCE_REQUIREMENTS_LIST'; Value : REG_RESOURCE_REQUIREMENTS_LIST));

  cMyRegistry     = 0;
  cClosedPage     = 1;
  cOpenPage       = 2;
  cClosedBook     = 3;
  cOpenBook       = 4;
  cDateValue      = 5;
  cStringValue    = 6;
  cBinaryValue    = 7;
  cAddBookmark    = 10;
  cRemoveBookmark = 11;

type
  TajRegReadWrite = (tRegRead, tRegWrite);

{--------------------------------------------------------------------------------------------------}
{                             Miscellaneous FileTime Functions                                     }
{--------------------------------------------------------------------------------------------------}

function FileTimeToDateTime(FileTime : TFileTime) : TDateTime;
// Convert FileTime to TDateTime.
var
  SystemTime  : TSystemTime;
begin
  FileTimeToLocalFileTime(FileTime, FileTime);
  FileTimeToSystemTime(FileTime, SystemTime);
  Result  := SystemTimeToDateTime(SystemTime);
end; {FileTimeToDateTime}

{--------------------------------------------------------------------------------------------------}
{                                 Registry - Read / Write Menu                                     }
{--------------------------------------------------------------------------------------------------}

procedure RegSubMenuSettings(Menu : TMenuItem; RegPath : string; ReadWrite : TajRegReadWrite);
var
  lp1   : integer;
  Name  : string;
begin
  with TRegistry.Create do begin
    try
      RootKey := cRegHKEY;
      if OpenKey(RegPath, ReadWrite = tRegWrite) then begin
        for lp1 := 0 to pred(Menu.Count) do begin
          if (Menu.Items[lp1].Count > 0) then
            RegSubMenuSettings(Menu.Items[lp1], RegPath, ReadWrite)
          else begin
            Name  := StripHotKey(Menu.Items[lp1].Caption);
            if (Name <> '-') then begin
              if (ReadWrite = tRegRead) then begin
                if ValueExists(Name) then begin
                  if (Menu.Items[lp1].Checked <> ReadBool(Name)) then
                    Menu.Items[lp1].Click;
                end; {if}
              end else
                WriteBool(Name, Menu.Items[lp1].Checked);
            end; {if}
          end; {if}
        end; {for}
        CloseKey;
      end; {if}
    finally;
      Free;
    end; {try}
  end; {with}
end; {RegSubMenuSettings}

{--------------------------------------------------------------------------------------------------}
{                             Miscellaneous Procedure & Functions                                  }
{--------------------------------------------------------------------------------------------------}

function GetNodePath(Node : TTreeNode) : string;
// Iterate from the node to the root building up the path string.
begin
  Result  := Node.Text;
  while (Node.Level > 1) do begin
    Node    := Node.Parent;
    Result  := Node.Text + '\' + Result;
  end; {while}
end; {GetNodePath}

{--------------------------------------------------------------------------------------------------}

procedure SetNodeImages(Node : TTreeNode; HasChildren : boolean);
begin
  if HasChildren then begin
    Node.HasChildren    := true;
    Node.ImageIndex     := cClosedBook;
    Node.SelectedIndex  := cOpenBook;
  end else begin
    Node.ImageIndex     := cClosedPage;
    Node.SelectedIndex  := cOpenPage;
  end; {if}
end; {SetNodeImages}

{--------------------------------------------------------------------------------------------------}
{                                     TajRegPathManager                                            }
{--------------------------------------------------------------------------------------------------}

procedure TajRegPathManager.SetfFullPath(Value : string);
var
  KeyText   : string;
  DelimPos  : integer;
begin
  if (fFullPath <> Value) then begin
    fFullPath := Value;
    DelimPos  := Pos('\', fFullPath);
    if (DelimPos > 0) then begin
      KeyText   := Copy(fFullPath, 1, pred(DelimPos));
      fSubPath  := Copy(fFullPath, succ(DelimPos), Length(fFullPath) - DelimPos);
    end else begin
      KeyText   := fFullPath;
      fSubPath  := '';
    end; {if}

    if (fRootHKEYText <> KeyText) then begin
      fRootHKEYText := KeyText;
      fRootHKEY     := HKEYTextToHKEY(fRootHKEYText);
    end; {if}

  end; {if}
end; {SetfFullPath}

{--------------------------------------------------------------------------------------------------}
{                                         TfrmRegBrowser                                           }
{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.FormCreate(Sender : TObject);
var
  lp1           : integer;
  Node1, Node2  : TTreeNode;
begin
  Width   := 632;
  Height  := 376;

  btnBookmark.Tag                 := cAddBookmark;                        // Initialize bookmark Tags.
  mnuPopBookmark.Tag              := cAddBookmark;

  IsPropertyEditor                := false;                               // Not a property editor - might be later.

  pnBack.DoubleBuffered           := true;                                // Stop some of the flicker.
  pnRegPath.DoubleBuffered        := true;
  edRegPath.DoubleBuffered        := true;

  fajAlterMenu                    := TajAlterMenu.Create(Self);
  fajRecentKeys                   := TajRecentKeys.Create(mnuBookmarks, fajAlterMenu, cRegRegistryBrowser);  // Add bookmark support.
  fajRecentKeys.OnRecentKeyClick  := RecentKeyClick;                      // Connect the event handler.

  fRegPathManager                 := TajRegPathManager.Create;            // Create the path manager.

  tvRegistry.Items.BeginUpdate;                                           // Add the default nodes ...
  Node1             := tvRegistry.Items.Add(nil, 'My Registry');          // HKEY_CURRENT_USER and its
  Node1.ImageIndex  := 0;                                                 // friends ...

  for lp1 := 1 to cNumRegHKEYKeys do begin
    Node2 := tvRegistry.Items.AddChild(Node1, cRegHKEYLookUp[lp1].Description);
    SetNodeImages(Node2, (RegNumSubKeys(cRegHKEYLookUp[lp1].Value, '') > 0));
  end; {for}

  Node1.Expand(false);                                                    // Expand the main root node.

  tvRegistry.Items.EndUpdate;
end; {FormCreate}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.FormShow(Sender : TObject);
begin
  RegSubMenuSettings(mnuView, cRegPath + cRegRegistryBrowser, tRegRead);
  if not fIsPropertyEditor then
    SetBounds((Screen.Width - Width) div 2, (Screen.Height - Height) div 2, Width, Height);
  btnCancel.Left  := pnToolbar.Width - btnCancel.Width - 3;
  btnUpdate.Left  := btnCancel.Left  - btnUpdate.Width - 3;
  edRegPath.Width := pnRegPath.Width - 8;
end; {FormShow}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.FormClose(Sender : TObject; var Action : TCloseAction);
begin
  RegSubMenuSettings(mnuView, cRegPath + cRegRegistryBrowser, tRegWrite);
end; {FormClose}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.FormDestroy(Sender : TObject);
begin
  fajAlterMenu.Free;
  fajRecentKeys.Free;                                                     // Free bookmark support plus
  fRegPathManager.Free;                                                   // the path manager.
end; {FormDestroy}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.SetfIsPropertyEditor(Value : boolean);
var
  lp1 : integer;
begin
  fIsPropertyEditor   := Value;                                           // Hide or reveal stuff that
  mnuUpdate.Visible   := fIsPropertyEditor;                               // relates to the usage as a
  mnuAddGUID.Visible  := fIsPropertyEditor;                               // property editor.
  for lp1 := 0 to pred(pnToolbar.ControlCount) do
    pnToolBar.Controls[lp1].Visible := (pnToolBar.Controls[lp1].Tag <> -1) or fIsPropertyEditor;
end; {SetfIsPropertyEditor}                                               // Uses tag of -1 for property editor stuff.

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.AddListNames(SubPath : string; List : TStringList);
type
  TRegBuffer  = record
    case integer of
      0 : (BuffChars  : array[0..1023] of char);
      1 : (BuffWord   : Cardinal);
    end;
var
  lp1, lp2    : integer;
  ListItem    : TListItem;
  Value       : string;
  LongWordVal : Cardinal;
  RegBuffer   : TRegBuffer;
  BuffType    : DWORD;
  BuffSize	  : DWORD;
  Key         : HKEY;

  {................................................................................................}

  procedure AddFileTimeInformation;
  var
    KeyFileTime : FileTime;
  begin
    if (RegQueryInfoKey(Key, nil, nil, nil, nil, nil, nil, nil, nil, nil, nil, @KeyFileTime) = ERROR_SUCCESS) then begin
      if (KeyFileTime.dwLowDateTime > 0) and (KeyFileTime.dwHighDateTime > 0) then begin
        ListItem            := lvRegistry.Items.Add;
        ListItem.ImageIndex := cDateValue;
        ListItem.Caption    := '(Modified)';
        ListItem.SubItems.Add('FILETIME');
        ListItem.SubItems.Add(FormatDateTime(ShortDateFormat + ' ' + 'hh:mm:ss:zzz', FileTimeToDateTime(KeyFileTime)));
      end; {if}
    end; {if}
  end; {AddFileTimeInformation}

  {................................................................................................}

begin
  with lvRegistry.Items do begin
    Clear;
    if (RegOpenKeyEx(fRegPathManager.RootHKEY, PChar(SubPath), 0, KEY_QUERY_VALUE, Key) = ERROR_SUCCESS) then begin
      List.Add('');
      List.Sort;

      AddFileTimeInformation;

      for lp1 := 0 to pred(List.Count) do begin
        Value     := '';
        ListItem  := Add;
        with ListItem do begin
          Caption     := List[lp1];
          ImageIndex  := cStringValue;
          BuffSize    := SizeOf(RegBuffer.BuffChars);
          if (RegQueryValueEx(Key, PChar(List[lp1]), nil, @BuffType, @(RegBuffer.BuffChars), @BuffSize) = ERROR_SUCCESS) then begin
            SubItems.Add(cRegValTypeLookUp[BuffType].Description);
            case BuffType of
              REG_NONE, REG_BINARY, REG_MULTI_SZ, REG_RESOURCE_LIST, REG_FULL_RESOURCE_DESCRIPTOR
                : begin
                    if (BuffSize = 0) then
                      Value := '(zero-length binary value)'
                    else begin
                      Value := '';
                      for lp2 := 0 to pred(BuffSize) do
                        Value := Value + IntToHex(ord(RegBuffer.BuffChars[lp2]), 2) + ' ';
                      Value := Trim(LowerCase(Value));
                    end; {if}
                    ImageIndex  := cBinaryValue;
                  end;
              REG_SZ, REG_EXPAND_SZ
                : Value := RegBuffer.BuffChars;
              REG_DWORD, REG_DWORD_BIG_ENDIAN
                : begin
                    LongWordVal := RegBuffer.BuffWord;
                    Value       := '0x' + LowerCase(IntToHex(LongWordVal, 8)) + ' (' + IntToStr(LongWordVal) + ')';
                    ImageIndex  := cBinaryValue;
                  end;
            else
              Value := 'ERROR - VALUE NOT READ';
            end; {case}
          end else begin
              SubItems.Add('');
              Value := '';
          end; {if}
          if (lp1 = 0) then begin
            Caption := '(Default)';
            if (Value = '') then
              Value := '(value not set)';
          end; {if}
          ListItem.SubItems.Add(Value);
        end; {with}
      end; {for}
    end; {if}
    RegCloseKey(Key);
  end; {with}
end; {AddListNames}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.AddTreeNodes(Node : TTreeNode; List : TStrings);
var
  lp1   : integer;
  Node1 : TTreeNode;
begin
  tvRegistry.Items.BeginUpdate;
  with tvRegistry.Items do begin
    for lp1 := 0 to pred(List.Count) do begin
      Node1 := AddChild(Node, List[lp1]);
      fRegPathManager.FullPath  := GetNodePath(Node1);
      SetNodeImages(Node1, (RegNumSubKeys(fRegPathManager.RootHKEY, fRegPathManager.SubPath)) > 0);
    end;
  end; {with}
  tvRegistry.Items.EndUpdate;
end; {AddTreeNodes}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.GetRegKeyValueNames(SubPath : string);
var
  KeyValueNameList  : TStringList;
begin
  KeyValueNameList        := TStringList.Create;
  KeyValueNameList.Sorted := true;
  KeyValueNameList.BeginUpdate;
  with TRegistry.Create do begin
    RootKey := fRegPathManager.RootHKEY;
    OpenKeyReadOnly(SubPath);
    GetValueNames(KeyValueNameList);
    CloseKey;
    Free;
  end; {with}
  KeyValueNameList.EndUpdate;
  AddListNames(SubPath, KeyValueNameList);
  KeyValueNameList.Free;
end; {GetRegKeyValueNames}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.GetRegKeyNames(SubPath : string; Node : TTreeNode);
var
  KeyNameList : TStringList;
begin
  KeyNameList         := TStringList.Create;
  KeyNameList.Sorted  := true;
  KeyNameList.BeginUpdate;
  with TRegistry.Create do begin
    RootKey := fRegPathManager.RootHKEY;
    OpenKeyReadOnly(SubPath);
    GetKeyNames(KeyNameList);
    CloseKey;
    Free;
  end; {with}
  KeyNameList.EndUpdate;  
  AddTreeNodes(Node, KeyNameList);
  KeyNameList.Free;
end; {GetRegKeyNames}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.SetBookmarkControls(Tag : integer);
var
  Glyph : TBitmap;
begin
  if (tvRegistry.Selected.Level > 1) then begin
    btnBookmark.Enabled     := true;
    mnuPopBookmark.Enabled  := true;
    if fajRecentKeys.InList(fRegPathManager.FullPath) then begin
      if (Tag = cAddBookmark) then begin
        Glyph := TBitmap.Create;
        imButtons.GetBitmap(1, Glyph);
        btnBookmark.Glyph.Assign(Glyph);
        Glyph.Free;
        btnBookmark.Hint          := 'Remove bookmark';
        btnBookmark.Tag           := cRemoveBookmark;
        mnuPopBookmark.ImageIndex := cRemoveBookmark;
        mnuPopBookmark.Caption    := '&Remove bookmark';
        mnuPopBookmark.Tag        := cRemoveBookmark;
      end; {if}
    end else begin
      if (Tag = cRemoveBookmark) then begin
        Glyph := TBitmap.Create;
        imButtons.GetBitmap(0, Glyph);
        btnBookmark.Glyph.Assign(Glyph);
        Glyph.Free;
        btnBookmark.Hint          := 'Add bookmark';
        btnBookmark.Tag           := cAddBookmark;
        mnuPopBookmark.ImageIndex := cAddBookmark;
        mnuPopBookmark.Caption    := '&Add bookmark';
        mnuPopBookmark.Tag        := cAddBookmark;
      end; {if}
    end; {if}
  end else begin
    btnBookmark.Enabled     := false;
    mnuPopBookmark.Enabled  := false;
  end; {if}
end; {SetBookmarkControls}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.tvRegistryChange(Sender : TObject; Node : TTreeNode);
begin
  if (Node.Level > 0) then begin
    lvRegistry.Items.BeginUpdate;
    tvRegistry.Update;
    fRegPathManager.FullPath  := GetNodePath(Node);
    edRegPath.Text            := fRegPathManager.FullPath;
    GetRegKeyValueNames(fRegPathManager.SubPath);
    lvRegistry.Items.EndUpdate;
  end else if (lvRegistry.Items.Count > 0) then begin
    lvRegistry.Items.BeginUpdate;
    lvRegistry.Items.Clear;
    lvRegistry.Items.EndUpdate;
  end {if};
  SetBookmarkControls(mnuPopBookmark.Tag);  
  tvRegistry.Update;  
  lvRegistry.Update;
end; {tvRegistryChange}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.tvRegistryExpanding(Sender : TObject; Node : TTreeNode; var AllowExpansion : boolean);
begin
  if (Node.Level > 0) then begin
    try
      fRegPathManager.FullPath  := GetNodePath(Node);

      if (Node.Level = 1) then
        Screen.Cursor := crHourGlass;

      if Node.HasChildren and (Node.GetFirstChild = nil) then
        GetRegKeyNames(fRegPathManager.SubPath, Node);
    finally
      Screen.Cursor := crDefault;
    end; {try}
  end; {if}
end; {tvRegistryExpanding}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.tvRegistryMouseDown(Sender : TObject; Button : TMouseButton; Shift : TShiftState; X, Y : integer);
begin
  if (Button = mbRight) then
    tvRegistry.Selected :=  tvRegistry.GetNodeAt(X, Y);
end; {tvRegistryMouseDown}

{--------------------------------------------------------------------------------------------------}

function TfrmRegBrowser.GoToBookmark(FullPath : string) : boolean;
var
  lp1       : integer;
  Node      : TTreeNode;
  LastNode  : TTreeNode;
  PathList  : TStringList;
  DelimPos  : integer;
  NodeFound : boolean;
  PathError : boolean;
begin
  PathList  := TStringList.Create;                            // Create our path list object.

  repeat                                                      // Fill the path list with the
    DelimPos  := Pos('\', FullPath);                          // parsed components of the requested path.
    if (DelimPos > 0) then begin
      PathList.Add(Copy(FullPath, 1, pred(DelimPos)));
      FullPath := Copy(FullPath, succ(DelimPos), Length(FullPath) - DelimPos);
    end else
      PathList.Add(FullPath);
  until (DelimPos = 0);

  PathError := false;                                         // Not started yet - so nothing's wrong.
  LastNode  := nil;                                           // Just in case nothing happens then this'll tell us.
  Node      := tvRegistry.Items[0];                           // Get the root node.
  lp1       := 0;                                             // Initialize our parsed path list counter.
  while not PathError and (lp1 < PathList.Count) do begin     // Keep on going until the end !
    Node  := Node.GetFirstChild;                              // Get the first child node of our node.
    if (Node <> nil) then                                     // Might hit the end stop !
      NodeFound := (Node.Text = PathList[lp1])                // Found our node ?
    else begin                                                // If it was a nil then we're never going
      NodeFound := false;                                     // to find the bookmarked node - so the
      PathError := true;                                      // path is bad !
    end; {if}
    while not NodeFound and not PathError do begin            // Scan all the remaining sibling nodes.
      Node  := Node.GetNextSibling;
      PathError := (Node = nil);                              // Keep an eye out for the end of the line !
      if not PathError then                                   // If not a bad path then see if we've got
        NodeFound := (Node.Text = PathList[lp1]);             // our baby !
    end; {while}

    inc(lp1);                                                 // Keep on going through our list.

    if (Node <> nil) then begin                               // If our last node isn't a node then
      if (lp1 < PathList.Count) then                          // expand our node but only if it's not
        Node.Expand(false);                                   // the last one in our list.
      LastNode  := Node;                                      // Remember this node as we'll need it for later.
    end; {if}

  end; {while}

  PathList.Free;                                              // Free the path list.

  if (LastNode <> nil) then begin
    LastNode.Selected         := true;                        // Select the last node - focus preserved somehow ?
    LastNode.MakeVisible;                                     // Make sure that we can see it too !
    fRegPathManager.FullPath  := GetNodePath(LastNode);       // Update the path to where we are now.
  end; {if}

  edRegPath.Text  := fRegPathManager.FullPath;                // Display the path in our edit box.

  if PathError then                                           // Oops - tell us about it !
    MessageDlg('Registry path error - key not found', mtWarning, [mbOK], 0);

  Result          := not PathError;                           // Did we do good ?

end; {GoToBookmark}

{--------------------------------------------------------------------------------------------------}

function TfrmRegBrowser.GetRegistryPath : string;
begin
  Result  := edRegPath.Text;
end; {GetRegistryPath}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.SetRegistryPath(Value : string);
begin
  AddKey(Value);
end; {SetRegistryPath}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.RecentKeyClick(Keyname : string);
begin
  Update;
  tvRegistry.Items.BeginUpdate;
  if GoToBookmark(Keyname) then
    fajRecentKeys.UpdateRecentKeys(Keyname, tAdd)
  else
    fajRecentKeys.UpdateRecentKeys(Keyname, tDelete);
  tvRegistry.Items.EndUpdate;
end; {RecentFileClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.DeleteBranchRoot(HKEYText : string);
var
  Node  : TTreeNode;
  Done  : boolean;
begin
  Node  := tvRegistry.Items[0].GetFirstChild;
  Done  := false;
  repeat
    if (Node <> nil) then begin
      if (Node.Text = HKEYText) then begin
        Node.DeleteChildren;
        Node.HasChildren  := (RegNumSubKeys(HKEYTextToHKEY(HKEYText), '') > 0);
        Done              := true;
      end; {if}
      Node  := Node.GetNextChild(Node);
    end else
      Done  := true;
  until Done;
end; {DeleteBranchRoot}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.AddKey(FullPath : string);
begin
  Update;
  tvRegistry.Items.BeginUpdate;
  fRegPathManager.FullPath  := FullPath;
  with TRegistry.Create do begin
    RootKey := fRegPathManager.RootHKEY;
    if not KeyExists(fRegPathManager.SubPath) then begin
      if (MessageDlg('Confirm creation of this key :' +#13#10#10 + fRegPathManager.FullPath, mtWarning, [mbYes, mbNo], 0) = mrYes) then begin
        OpenKey(fRegPathManager.SubPath, true);
        CloseKey;
        DeleteBranchRoot(fRegPathManager.RootHKEYText);
      end; {if}
    end; {if}
    Free;
    GoToBookmark(fRegPathManager.FullPath);
  end; {with}
  tvRegistry.Items.EndUpdate;
end; {AddKey}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.DeleteKey(FullPath : string);
var
  SubPath       : string;
  KeyName       : string;
  KeyNameList   : TStringList;
  KeyNameIndex  : integer;
begin
  Update;
  tvRegistry.Items.BeginUpdate;
  fRegPathManager.FullPath  := FullPath;
  with TRegistry.Create do begin
    RootKey := fRegPathManager.RootHKEY;
    if KeyExists(fRegPathManager.SubPath) then begin
      if (RegNumSubKeys(fRegPathManager.RootHKEY, fRegPathManager.SubPath) > 0) then
        MessageDlg('Can not delete populated node :' +#13#10#10 +
                   fRegPathManager.FullPath, mtError, [mbOK], 0)
      else if (MessageDlg('Confirm deletion of this key :' +#13#10#10 +
                          fRegPathManager.FullPath, mtWarning, mbOKCancel, 0) = mrOK) then begin
        SubPath := fRegPathManager.SubPath;
        KeyName := ExtractFileName(SubPath);

        fRegPathManager.FullPath  := ExcludeTrailingBackSlash(ExtractFilePath(fRegPathManager.FullPath));

        KeyNameList         := TStringList.Create;
        KeyNameList.Sorted  := true;
        KeyNameList.BeginUpdate;
        OpenKeyReadOnly(fRegPathManager.SubPath);
        GetKeyNames(KeyNameList);
        CloseKey;
        KeyNameList.EndUpdate;

        if (KeyNameList.Count > 1) then begin
          KeyNameIndex := KeyNameList.IndexOf(KeyName);
          if (KeyNameIndex > 0) then
            KeyName := KeyNameList[pred(KeyNameIndex)]
          else
            KeyName := KeyNameList[succ(KeyNameIndex)];
          fRegPathManager.FullPath := fRegPathManager.FullPath + '\' + KeyName;
        end; {if}

        KeyNameList.Free;

        DeleteKey(SubPath);

        DeleteBranchRoot(fRegPathManager.RootHKEYText);

      end; {if}
    end; {if}

    Free;
    GoToBookmark(fRegPathManager.FullPath);

  end; {with}
  tvRegistry.Items.EndUpdate;
end; {DeleteKey}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.lvPopupMenuPopup(Sender : TObject);
begin
  if (lvRegistry.Selected <> nil) then begin
    with lvRegistry.Selected do
      mnuDeleteValue.Enabled  := (Caption <> '(Default)') and (Caption <> '(Modified)');
  end else
    mnuDeleteValue.Enabled  := false;
  mnuEditValue.Enabled  := mnuDeleteValue.Enabled;
end; {lvPopupMenuPopup}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.DeleteValue(KeyValue : string);
begin
  if (MessageDlg('Confirm deletion of this value : ' + KeyValue + #13#10#10 + 'From :' + #13#10#10 +
                  fRegPathManager.FullPath,
                  mtWarning, mbOKCancel, 0) = mrOK) then begin
    with TRegistry.Create do begin
      RootKey := fRegPathManager.RootHKEY;
      OpenKey(fRegPathManager.SubPath, false);
      if DeleteValue(KeyValue) then
        lvRegistry.Selected.Delete;
      CloseKey;
      Free;
    end; {with}
  end; {if}
end; {DeleteValue}

{--------------------------------------------------------------------------------------------------}
{                                     Menu Event Handlers                                          }
{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuAboutClick(Sender : TObject);
// What else ? - it show the about box !
begin
  with TfrmAbout.Create(nil) do begin
    try
      if fIsPropertyEditor then
        Caption  := 'Key Maker Package'
      else
        Caption  := 'Registry Browser';
      ShowModal;
    finally
      Free;
    end; {try}
  end; {with}
end; {mnuAboutClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuAddGUIDClick(Sender : TObject);
// Adds a GUID to the current path string - only used in the property editor - just to deceive hackers.
var
  GUID  : TGUID;
begin
  if (CoCreateGuid(GUID) = S_OK) then
    edRegPath.Text  := IncludeTrailingBackslash(edRegPath.Text) + GUIDToString(GUID);
end; {mnuAddGUIDClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuAddKeyClick(Sender : TObject);
// Go and add a new key that's specified by edRegPath.Text.
begin
  AddKey(edRegPath.Text);
end; {mnuAddKeyClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuBookmarkClick(Sender : TObject);
// Add or delete bookmarks - selection of which is done by the Sender's Tag property.
begin
  if (TWinControl(Sender).Tag = cAddBookmark) then
    fajRecentKeys.UpdateRecentKeys(fRegPathManager.FullPath, tAdd)
  else
    fajRecentKeys.UpdateRecentKeys(fRegPathManager.FullPath, tDelete);
  SetBookmarkControls(TWinControl(Sender).Tag);
end; {mnuBookmarkClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuDeleteKeyClick(Sender : TObject);
// Delete the key specfied by edRegPath.Text.
begin
  DeleteKey(edRegPath.Text);
end; {mnuDeleteKeyClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuDeleteValueClick(Sender : TObject);
// Delete the key value.
begin
  if (lvRegistry.Selected <> nil) then
    DeleteValue(lvRegistry.Selected.Caption);
end; {mnuDeleteValueClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuExitClick(Sender : TObject);
// Looks like we're leaving !
begin
  if fIsPropertyEditor then
    ModalResult := mrCancel
  else
    Close;
end; {mnuExitClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuRefreshClick(Sender : TObject);
// Refresh the branch that we're looking at and close all others.
var
  Node  : TTreeNode;
begin
  tvRegistry.Items.BeginUpdate;
  if (tvRegistry.Selected.Level > 0) then
    fRegPathManager.FullPath  := GetNodePath(tvRegistry.Selected);
  Node  := tvRegistry.Items[0].GetFirstChild;
  repeat
    Node.DeleteChildren;
    DeleteBranchRoot(Node.Text);
    if (Node.Text = fRegPathManager.RootHKEYText) and (fRegPathManager.SubPath <> '') then
      GoToBookmark(fRegPathManager.FullPath);
    Node  := Node.GetNextChild(Node);
  until (Node = nil);
  tvRegistry.Items.EndUpdate;
end; {mnuRefreshClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuUpdateClick(Sender: TObject);
// Closing time but tell the property manager that it's OK to update the new value.
begin
  if fIsPropertyEditor then
    ModalResult := mrOK
  else
    Close;
end; {mnuUpdateClick}

{--------------------------------------------------------------------------------------------------}

procedure TfrmRegBrowser.mnuXPMenuClick(Sender : TObject);
// Set XP menu styles.
begin
  with Sender as TMenuItem do begin
    Checked               := not Checked;
    fajAlterMenu.Enabled  := Checked;
  end; {with}
end; {mnuXPMenuClick}

{--------------------------------------------------------------------------------------------------}
{ajRegistryBrowser}
end.
