          ///////////////////////////////////////////////////////////////////////
          //                                                                   //
          //       SoftSpot Software Component Library                         //
          //       Key Maker software component                                //
          //                                                                   //
          //       Copyright (c) 1996 - 2002 SoftSpot Software Ltd.            //
          //       ALL RIGHTS RESERVED                                         //
          //                                                                   //
          //   The entire contents of this file is protected by U.S. and       //
          //   International Copyright Laws. Unauthorized reproduction,        //
          //   reverse-engineering, and distribution of all or any portion of  //
          //   the code contained in this file is strictly prohibited and may  //
          //   result in severe civil and criminal penalties and will be       //
          //   prosecuted to the maximum extent possible under the law.        //
          //                                                                   //
          //   RESTRICTIONS                                                    //
          //                                                                   //
          //   THE SOURCE CODE CONTAINED WITHIN THIS FILE AND ALL RELATED      //
          //   FILES OR ANY PORTION OF ITS CONTENTS SHALL AT NO TIME BE        //
          //   COPIED, TRANSFERRED, SOLD, DISTRIBUTED, OR OTHERWISE MADE       //
          //   AVAILABLE TO OTHER INDIVIDUALS WITHOUT EXPRESS WRITTEN CONSENT  //
          //   AND PERMISSION FROM SOFTSPOT SOFTWARE LTD.                      //
          //                                                                   //
          //   CONSULT THE END USER LICENSE AGREEMENT FOR INFORMATION ON       //
          //   ADDITIONAL RESTRICTIONS.                                        //
          //                                                                   //
          ///////////////////////////////////////////////////////////////////////

unit ajTimeEdit;

{                           -=< Key Maker File Browser Time Editor >=-
{
{ Copyright SoftSpot Software 2002 - All Rights Reserved
{
{ Author        : Andrew J Jameson
{
{ Web site      : www.softspotsoftware.com
{ e-mail        : contact@softspotsoftware.com
{
{ Creation Date : 01 March 2002
{
{ Version       : 1.00
{
{ Description   : Time editor component for Key Maker - edits milliseconds too !                   }

interface

uses
  StdCtrls, Messages, Controls, Classes, Windows, ExtCtrls;

type
  TajButtonType   = (tUpButton, tDownButton);

type
  TajButton = class(TObject)
  private
    fOwner      : TWinControl;                                            // Needed to get the owner's canvas.
    fButtonType : TajButtonType;                                          // Well it's either an up or down button.
    fDown       : boolean;                                                // And that button can be eith up or down.
    fRect       : TRect;                                                  // The button's bounds rect.
    fHasCapture : boolean;                                                // Hey ... we captured the mouse !
  protected
    procedure   SetfDown    (Value : boolean);                            // Set the button - up or down.
  public
    constructor Create      (AOwner : TWinControl; ButtonType : TajButtonType);
    procedure   Draw;
    property    ButtonType  : TajButtonType   read fButtonType    write fButtonType;
    property    Down        : boolean         read fDown          write SetfDown;
    property    HasCapture  : boolean         read fHasCapture    write fHasCapture;
    property    Bounds      : TRect           read fRect          write fRect;
  end;

type
  TajTimeEdit = class(TCustomEdit)
  private
    { Private declarations }
    fSelIndex       : integer;                                            // Index of the selected field - hr, min , sec or mSecs.
    fOldWindowProc  : TWndMethod;                                         // The TEdit's old WindowProc.
    fTimeChange     : TNotifyEvent;                                       // Time changed notify event.
    fDateTime       : TDateTime;                                          // The actual time value.
    fUpButton       : TajButton;                                          // The up button and it's brother ...
    fDownButton     : TajButton;                                          // the down button.
    fTimer          : TTimer;                                             // Auto repeat timer.
    procedure   NewWindowProc   (var Message : TMessage);                 // The new WindowProc.
    procedure   pWMChar         (Message : TWMChar);
    procedure   pWMKeyDown      (Message : TWMChar);
    procedure   pWMLButtonUp    (Message : TWMMouse);
    procedure   pWMMove         (Message : TWMMouse);
    procedure   pWMLButtonClick (Message : TWMMouse);
    procedure   pWMLButtonDown  (Message : TWMMouse);
    procedure   pWMPaint        (Message : TMessage);
    procedure   SetSelValue     (Value : integer);
    procedure   RepeatClick     (Sender : TObject);
  protected
    procedure   CreateWnd;                                override;
    function    GetTime         : TDateTime;
    procedure   SetTime         (Value : TDateTime);
  public
    { Public declarations }
    constructor Create          (AOwner : TComponent);    override;
    destructor  Destroy;                                  override;
    procedure   ClearTime;
    property    DateTime        : TDateTime     read GetTime      write SetTime;
    property    OnTimeChange    : TNotifyEvent  read fTimeChange  write fTimeChange;
  end;

implementation

uses
  SysUtils, Graphics;

type
  TSelPos = record                                                        // Used to set the selected
    SelStart  : integer;                                                  // text of the TEdit.
    SelLength : integer;
  end;

const
  cSelHours = 0;
  cSelMins  = 1;
  cSelSecs  = 2;
  cSelmSecs = 3;
  cSelText  : array[cSelHours..cSelmSecs] of TSelPos =
                                        ((SelStart : 0; SelLength : 2),   // The selected field constants
                                         (SelStart : 3; SelLength : 2),   // for hr, mins, secs and mSecs.
                                         (SelStart : 6; SelLength : 2),
                                         (SelStart : 9; SelLength : 3));

  cButtonWidth  = 17;                                                     // Guess !

{--------------------------------------------------------------------------------------------------}
{                                       TajButton                                                  }
{--------------------------------------------------------------------------------------------------}

constructor TajButton.Create(AOwner : TWinControl; ButtonType : TajButtonType);
begin
  fOwner      := AOwner;                                                  // Store the owner.
  fButtonType := ButtonType;                                              // I'm either an up or down button.
  fDown       := false;                                                   // And I'm not pressed.
  fHasCapture := false;                                                   // And I've not got the mouse yet.
end; {constructor}

{--------------------------------------------------------------------------------------------------}

procedure TajButton.SetfDown(Value : boolean);
begin
  if (fDown <> Value) then begin                                          // Looks as though I'm going up or
    fDown := Value;                                                       // down in the world ... so redraw
    Draw;                                                                 // me.
  end; {if}
end; {SetfDown}

{--------------------------------------------------------------------------------------------------}

procedure TajButton.Draw;
var
  cx, cy  : integer;
  aw, ah  : integer;
  Arrow   : array [0..2] of TPoint;
  DC      : HDC;
  lp1     : integer;
begin
  fRect := Rect(1, 0, cButtonWidth, fOwner.ClientHeight div 2);
  if (fButtonType = tUpButton) then begin
    OffSetRect(fRect, fOwner.ClientWidth - cButtonWidth, 0);
    with fRect do begin
      cx  := Left + cButtonWidth   div 2;
      cy  := Top  + (Bottom - Top) div 2;
      if fDown or not fOwner.Enabled then begin
        inc(cx);
        inc(cy);
      end; {if}
      if (cButtonWidth > Bottom - Top) then begin
        ah  := (Bottom - Top) div 2 - 2;
        aw  := ah * 2 - 1;
      end else begin
        aw  := cButtonWidth div 2 - 1;
        ah  := aw div 2 + 1;
      end; {if}
      Arrow[0]  := Point(cx,            cy - ah div 2 + 1);
      Arrow[1]  := Point(cx - aw div 2, cy + ah div 2    );
      Arrow[2]  := Point(cx + aw div 2, cy + ah div 2    );
    end; {with}
  end else begin
    OffSetRect(fRect, fOwner.ClientWidth - cButtonWidth, fOwner.ClientHeight div 2 + 1);
    with fRect do begin
      cx  := Left + cButtonWidth   div 2;
      cy  := Top  + (Bottom - Top) div 2;
      if fDown or not fOwner.Enabled then begin
        inc(cx);
        inc(cy);
      end; {if}
      if (cButtonWidth > Bottom - Top) then begin
        ah  := (Bottom - Top) div 2 - 2;
        aw  := ah * 2 - 1;
      end else begin
        aw  := cButtonWidth div 2 - 1;
        ah  := aw div 2 + 1;
      end; {if}
      Arrow[0]  := Point(cx - aw div 2, cy - ah div 2    );
      Arrow[1]  := Point(cx + aw div 2, cy - ah div 2    );
      Arrow[2]  := Point(cx,            cy + ah div 2 - 1);
    end; {with}
  end; {if}

  DC  := GetDC(fOwner.Handle);

  with TCanvas.Create, fRect do begin
    Handle      := DC;
    Brush.Color := clBtnFace;
    FillRect(fRect);
    if fOwner.Enabled then begin
      Brush.Color := clBlack;
      Pen.Color   := clBlack;
      Polygon(Arrow);
    end else begin
      Brush.Color := clBtnHighlight;
      Pen.Color   := clBtnHighlight;
      Polygon(Arrow);
      for lp1 := 0 to 2 do begin
        Arrow[lp1].x  := Arrow[lp1].x - 1;
        Arrow[lp1].y  := Arrow[lp1].y - 1;
      end; {for}
      Brush.Color := clBtnShadow;
      Pen.Color   := clBtnShadow;
      Polygon(Arrow);
    end; {if}
    if fDown then begin
      Pen.Color := clBtnShadow;
      Polyline([Point(Left,   Bottom-1), Point(Left,   Top  ), Point(Right,   Top  )]);
      Pen.Color := clBlack;
      Polyline([Point(Left+1, Bottom-2), Point(Left+1, Top+1), Point(Right-1, Top+1)]);
    end else begin
      Pen.Color := clBtnHighlight;
      Polyline([Point(Left+1, Bottom-2), Point(Left+1,  Top+1   ), Point(Right-1, Top+1)]);
      Pen.color := clBtnShadow;
      Polyline([Point(Left+1, Bottom-1), Point(Right-1, Bottom-1), Point(Right-1, Top  )]);
    end; {if}
    if (fButtonType = tUpButton) then begin
      if fDown then
        Pen.Color := clBtnHighlight
      else
        Pen.Color := clBlack;
      MoveTo(fOwner.ClientWidth - cButtonWidth + 1, fOwner.ClientHeight div 2);
        LineTo(fOwner.ClientWidth, fOwner.ClientHeight div 2);
    end; {if}
    ReleaseDC(Handle, DC);
    Free;
  end; {with}
end; {DrawButton}

{--------------------------------------------------------------------------------------------------}
{                                       TajTimeEdit                                                }
{--------------------------------------------------------------------------------------------------}

constructor TajTimeEdit.Create(AOwner : TComponent);
begin
  inherited;
  fSelIndex       := cSelHours;                                           // Initial selected field is hours.
  fDateTime       := Now;                                                 // Set the time (and date) to now.
  fUpButton       := TajButton.Create(Self, tUpButton  );                 // Create the up button.
  fDownButton     := TajButton.Create(Self, tDownButton);                 // Create the down button.
  fTimer          := TTimer.Create(Self);                                 // Create the auto repeat timer.
  fTimer.Enabled  := false;                                               // Don't need it yet.
  fTimer.Interval := 500;                                                 // Set the initial delay to 0.5 secs.
  fTimer.OnTimer  := RepeatClick;                                         // Connect its event handler.
  fOldWindowProc  := WindowProc;                                          // Keep a copy of the old WindowProc ...
  WindowProc      := NewWindowProc;                                       // and give it a new one.
  Text            := '00:00:00:000';                                      // Set the text to its default.
end; {constructor}

{--------------------------------------------------------------------------------------------------}

destructor TajTimeEdit.Destroy;
begin
  WindowProc  := fOldWindowProc;                                          // Restore the old WindowProc.
  fUpButton.Free;                                                         // Free the buttons.
  fDownButton.Free;
  inherited;
end; {destructor}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.CreateWnd;
begin
  inherited;
  Perform(EM_SETMARGINS, EC_RIGHTMARGIN, (cButtonWidth + 2) shl 16);      // Reserve space for the buttons.
end; {CreateWnd}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.NewWindowProc(var Message : TMessage);
begin
  case Message.Msg of
    WM_CHAR           : pWMChar(TWMKey(Message));
    WM_KEYDOWN        : pWMKeyDown(TWMKey(Message));
    WM_LBUTTONUP      : pWMLButtonUp(TWMMouse(Message));
    WM_MOUSEMOVE      : pWMMove(TWMMouse(Message));
    WM_LBUTTONDOWN    : pWMLButtonDown(TWMMouse(Message));
    WM_LBUTTONDBLCLK  : pWMLButtonClick(TWMMouse(Message));
    WM_PAINT          : pWMPaint(Message);
    WM_RBUTTONDOWN,
    WM_SETCURSOR      : ;                                                 // Swallow these ... we don't want them !
  else begin
      if (Message.Msg = WM_NCHITTEST) then
        HideCaret(0);
      if (Message.Msg = EM_SETSEL) then begin                             // Intercept and set our selected text.
        Message.WParam  := cSelText[fSelIndex].SelStart;
        Message.LParam  := cSelText[fSelIndex].SelStart + cSelText[fSelIndex].SelLength;
      end; {if}
      fOldWindowProc(Message);                                            // Pass these and others onward.
    end; {if}
  end; {case}
end; {NewWindowProc}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.pWMLButtonUp(Message : TWMMouse);
begin
  ReleaseCapture;                                                         // Release mouse capture.
  fUpButton.Down          := false;                                       // Set button up - redraws if it was down.
  fUpButton.HasCapture    := false;                                       // Button no longer has capture.
  fDownButton.Down        := false;                                       // Set button up - redraws if it was down.
  fDownButton.HasCapture  := false;                                       // Button no longer has capture.
  fTimer.Enabled          := false;                                       // Turn the timer off.
  fTimer.Interval         := 500;                                         // Reset the intial interval to 0.5 secs.
end; {pWMLButtonUp}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.pWMMove(Message : TWMMouse);
begin
  fUpButton.Down    := fUpButton.HasCapture   and PtInRect(fUpButton.Bounds,   Point(Message.XPos, Message.YPos));
  fDownButton.Down  := fDownButton.HasCapture and PtInRect(fDownButton.Bounds, Point(Message.XPos, Message.YPos));
end; {pWMMove}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.RepeatClick(Sender : TObject);
begin
  if fUpButton.Down then
    SetSelValue(StrToInt(SelText) + 1)
  else if fDownButton.Down then
    SetSelValue(StrToInt(SelText) - 1);
  fTimer.Interval := 100;                                                 // Make the interval shorter.
end; {RepeatClick}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.pWMLButtonDown(Message : TWMMouse);
begin

  SetFocus;
  SetCapture(Handle);

  SelStart  := cSelText[fSelIndex].SelStart;
  SelLength := cSelText[fSelIndex].SelLength;

  with Message do begin
    if PtInRect(fUpButton.Bounds, Point(XPos, YPos)) then begin

      fUpButton.Down          := true;
      fUpButton.HasCapture    := true;
      fTimer.Enabled          := true;
      SetSelValue(StrToInt(SelText) + 1);

    end else if PtInRect(fDownButton.Bounds, Point(XPos, YPos)) then begin

      fDownButton.Down        := true;
      fDownButton.HasCapture  := true;
      fTimer.Enabled          := true;
      SetSelValue(StrToInt(SelText) - 1);

    end else if (Msg <> WM_LBUTTONDBLCLK) and                             // Don't come in here clicking away ...
      (XPos < ClientWidth - cButtonWidth) then begin                      // Ignore the pixel high gap between the buttons.

      fOldWindowProc(TMessage(Message));                                  // Need to let the TEdit position the SelStart.

      case SelStart of
        0..2  : fSelIndex := cSelHours;
        3..5  : fSelIndex := cSelMins;
        6..8  : fSelIndex := cSelSecs;
        9..12 : fSelIndex := cSelmSecs;
      end; {case}

      SelStart  := cSelText[fSelIndex].SelStart;
      SelLength := cSelText[fSelIndex].SelLength;

    end; {if}
  end; {with}

end; {pWMLButtonDown}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.pWMLButtonClick(Message : TWMMouse);
begin
  pWMLButtonUp  (Message);                                                // Forgot clicks ... do up and
  pWMLButtonDown(Message);                                                // then down ... seems the wrong
end; {pWMLButtonClick}                                                    // way round - but it works.

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.pWMPaint(Message : TMessage);
begin
  fOldWindowProc(Message);                                                // Let it draw what it needs to.
  fUpButton.Draw;                                                         // Now draw the buttons.
  fDownButton.Draw;
end; {pWMPaint}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.SetSelValue(Value : integer);
var
  NumText : string;
begin
  case fSelIndex of
    cSelHours           : if (Value < 0) then
                            Value := Value + 24
                          else if (Value > 23) then
                            Value := Value - (Value div 10) * 10;         // Too big, so just use the units.

    cSelMins, cSelSecs  : if (Value < 0) then
                            Value := Value + 60
                          else if (Value > 59) then
                            Value := Value - (Value div 10) * 10;         // Too big, so just use the units.

    cSelmSecs           : if (Value < 0) then
                            Value := Value + 1000
                          else if (Value > 999) then
                            Value := Value - 1000;
  end; {case}
  NumText   := IntToStr(Value);
  while (Length(NumText) < cSelText[fSelIndex].SelLength) do
    NumText := '0' + NumText;
  SelText   := NumText;
  SelStart  := cSelText[fSelIndex].SelStart;
  SelLength := cSelText[fSelIndex].SelLength;
  if Assigned(fTimeChange) then                                           // Fire the time change event if assigned.
    fTimeChange(Self);
end; {SetSelValue}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.pWMChar(Message : TWMChar);
begin
  if chr(Message.CharCode) in ['0'..'9'] then
    SetSelValue(StrToInt(Copy(SelText + chr(Message.CharCode), 2, Length(SelText))));
end; {pWMChar}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.pWMKeyDown(Message : TWMChar);
begin
  case Message.CharCode of
    VK_LEFT   : fSelIndex := (fSelIndex + 3) mod 4;
    VK_RIGHT  : fSelIndex := (fSelIndex + 1) mod 4;
    VK_UP     : SetSelValue(StrToInt(SelText) + 1);
    VK_DOWN   : SetSelValue(StrToInt(SelText) - 1);
  end; {case}
  SelStart  := cSelText[fSelIndex].SelStart;
  SelLength := cSelText[fSelIndex].SelLength;
end; {pWMKeyDown}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.ClearTime;
begin
  Text  := '00:00:00:000';
end; {ClearTime}

{--------------------------------------------------------------------------------------------------}

function TajTimeEdit.GetTime : TDateTime;
var
  Hour, Mins, Secs, MSecs : Word;
begin
  Hour    := StrToInt(Copy(Text,  1, 2));
  Mins    := StrToInt(Copy(Text,  4, 2));
  Secs    := StrToInt(Copy(Text,  7, 2));
  MSecs   := StrToInt(Copy(Text, 10, 3));
  Result  := Trunc(fDateTime) + EncodeTime(Hour, Mins, Secs, MSecs);
end; {GetTime}

{--------------------------------------------------------------------------------------------------}

procedure TajTimeEdit.SetTime(Value : TDateTime);
begin
  fDateTime := Value;
  Text      := FormatDateTime('hh:mm:ss:zzz', Value);
end; {SetTime}

{--------------------------------------------------------------------------------------------------}
{ajTimeEdit}
end.
